###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handler that executes { "hello":{...}} task in recipe JSON.

This is meant as an example only, it executes no useful tasks. Use this as
a template for how to connect a handler to a JSON recipe task.  It
illustrates how to use JOSN variables, access constants in the system, and
best practices for using this framework.  Credentials are not required for
this recipe handler.

Call from the command line using:

  - `python all/run.py scripts/say_hello.json`

Call from code using:

  - `hello(recipe, instance)`

"""

import json
import time

from starthinker.util.project import project


@project.from_parameters
def hello():

  if project.verbose:
    print('HELLO')

  print('')
  print('-' * 80)
  print('Tasks are just python, you can do whatever.')
  print('This is a task being executed.')
  print('')
  print('SAY:', project.task['say'])
  print('')
  print('')

  print('-' * 80)
  print('Every task automatically gets a date parameter.')
  print('The parameter can be passed in, or defaults to today.')
  print('')
  print('PROJECT DATE:', project.date)
  print('')
  print('')

  print('-' * 80)
  print('Most tasks operate on top of Google Cloud infrastructure.')
  print("Every task specifies an 'auth' parameter as 'user' or 'service'.")
  print('Every project has its own credentails paths.')
  print(
      "If you provide 'client' credentials, the 'user' credentials will be populated as necessary."
  )
  print(
      "If you provde the 'user' credentails, the 'client' credentials are not necessary."
  )
  print("If you use the 'service' credentials, you must add them manually.")
  print('')
  print('PROJECT ID:', project.id)
  print(
      'PROJECT CLIENT CREDENTIALS:',
      project.recipe.get('setup', {}).get('auth', {}).get('client', 'MISSING'))
  print('PROJECT USER CREDENTIALS:',
        project.recipe.get('setup', {}).get('auth', {}).get('user', 'MISSING'))
  print(
      'PROJECT SERVICE CREDENTIALS:',
      project.recipe.get('setup', {}).get('auth', {}).get('service', 'MISSING'))
  print('')
  print('')

  print('-' * 80)
  print('Your entire project definition is accessible as a dictionary.')
  print('The task name must match a directory with a run.py inside it.')
  print(
      "For example, 'hello' is a task which will executed by 'task/hello/run.py'."
  )
  print('')
  print('PROJECT JSON:')
  print(json.dumps(project.recipe, indent=2, default=str))

  print('')
  print('')

  print('-' * 80)
  print('Each task is passed a nested subset of json.')
  print(
      'Different tasks should NOT share json. Security and readability reasons.'
  )
  print('Each task can execute as a service or a user independently.')
  print('Access structure data within a task as...')
  print('')
  print('PROJECT TASK:', project.task)
  print('PROJECT TASK AUTH:', project.task['auth'])
  print('PROJECT TASK SAY:', project.task['say'])
  print('')
  print('')

  if project.task.get('sleep'):
    print('PROJECT TASK SLEEP:', project.task['sleep'])
    time.sleep(project.task['sleep'])

  if project.task.get('error'):
    print('PROJECT TASK ERROR TRIGGERED:')
    raise Exception(project.task['error'])

  print('-' * 80)
  print("Take a look inside 'task/hello/run.py'.")
  print('Its a great skeleton for your first project.')
  print('')
  print('')


if __name__ == '__main__':
  hello()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################


from starthinker.util.bigquery import datasets_access
from starthinker.util.bigquery import datasets_create
from starthinker.util.bigquery import datasets_delete
from starthinker.util.project import project


@project.from_parameters
def dataset():
  if project.verbose:
    print('DATASET', project.id, project.task['dataset'])

  if project.task.get('delete', False):
    if project.verbose:
      print('DATASET DELETE')
    # In order to fully delete a dataset, it needs to first have all tables
    # deleted, which is done with the delete_contents=True, and then the actual
    # dataset can be deleted, which is done with delete_contents=false.
    datasets_delete(
        project.task['auth'],
        project.id,
        project.task['dataset'],
        delete_contents=True
    )
    datasets_delete(
        project.task['auth'],
        project.id,
        project.task['dataset'],
        delete_contents=False
    )
  else:
    if project.task.get('clear', False):
      if project.verbose:
        print('DATASET CLEAR')
      datasets_delete(
          project.task['auth'],
          project.id,
          project.task['dataset'],
          delete_contents=True
      )

    if project.verbose:
      print('DATASET CREATE')
    datasets_create(
        project.task['auth'],
        project.id,
        project.task['dataset']
    )

    if project.verbose:
      print('DATASET ACCESS')
    datasets_access(
        project.task['auth'],
        project.id,
        project.task['dataset'],
        emails=project.task.get('emails', []),
        groups=project.task.get('groups', [])
    )


if __name__ == '__main__':
  dataset()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of Ads."""

import time, json

from starthinker.task.traffic.campaign import CampaignDAO
from starthinker.task.traffic.class_extensions import StringExtensions
from starthinker.task.traffic.creative import CreativeDAO
from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.event_tag import EventTagDAO
from starthinker.task.traffic.feed import FieldMap
from starthinker.task.traffic.landing_page import LandingPageDAO
from starthinker.task.traffic.placement import PlacementDAO
from starthinker.task.traffic.store import store


class AdDAO(BaseDAO):
  """Ad data access object.

  Inherits from BaseDAO and implements ad specific logic for creating and
  updating ads.
  """

  def __init__(self, auth, profile_id, is_admin):
    """Initializes AdDAO with profile id and authentication scheme."""
    super(AdDAO, self).__init__(auth, profile_id, is_admin)

    self._id_field = FieldMap.AD_ID
    self._search_field = FieldMap.AD_NAME
    self._list_name = 'ads'

    self._creative_dao = CreativeDAO(auth, profile_id, is_admin)
    self._placement_dao = PlacementDAO(auth, profile_id, is_admin)
    self._campaign_dao = CampaignDAO(auth, profile_id, is_admin)
    self._event_tag_dao = EventTagDAO(auth, profile_id, is_admin)
    self._landing_page_dao = LandingPageDAO(auth, profile_id, is_admin)

    self._parent_filter_name = 'campaignIds'
    self._parent_dao = self._campaign_dao
    self._parent_filter_field_name = FieldMap.CAMPAIGN_ID

    self._entity = 'AD'

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(AdDAO, self)._api(iterate).ads()

  def _api_creatives(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(AdDAO, self)._api(iterate).creatives()

  def _wait_creative_activation(self, creative_id, timeout=128):
    """Waits for a creative to become active.

    This function checks the if the creative is active in intervals that
    increase exponentially (exponential backoff).

    Args:
      creative_id: Creative identifier.
      timeout: Optional parameter, determines how many seconds to wait for the
        activation.

    Raises:
      Exception: In case the creative doesn't activate within the specified
      timeout

    """
    # Only wait for creative activation if it is a new creative trafficked by
    # this Bulkdozer session
    if store.get('CREATIVE', creative_id):
      creative = self._api_creatives().get(
          profileId=self.profile_id, id=creative_id).execute()
      wait = 2

      while not creative['active'] and timeout > 0:
        print('Waiting %s seconds for creative %s activation...' %
              (wait, creative_id))
        time.sleep(wait)
        timeout -= wait
        wait *= 2
        creative = self._api_creatives().get(
            profileId=self.profile_id, id=creative_id).execute()

      if not creative['active']:
        raise Exception(
            'Creative %s failed to activate within defined timeout' %
            creative['id'])

  def _wait_all_creative_activation(self, feed_item, timeout=128):
    """Waits for activation of all creatives that should be associated to the feed item that represents an ad.

    Args:
      feed_item: Feed item representing an Ad from the Bulkdozer feed.
      timeout: Optional parameter identifying how long to wait for all creatives
        to be activated in seconds.

    Raises:
      Exception: In case one or more creatives do not get activated within the
      specified timeout.

    """
    for association in feed_item['creative_assignment']:
      creative = self._creative_dao.get(association, required=True)
      self._wait_creative_activation(creative['id'], timeout)

  def _assignment_matches(self, item, assignment):
    if item.get(FieldMap.AD_ID, None) and assignment.get(FieldMap.AD_ID, None):
      return item.get(FieldMap.AD_ID,
                      None) == assignment.get(FieldMap.AD_ID, None)
    else:
      return item.get(FieldMap.AD_NAME,
                      '1') == assignment.get(FieldMap.AD_NAME, '2')

  def map_feeds(self, ad_feed, ad_creative_assignment, ad_placement_assignment,
                ad_event_tag_assignment, placement_feed,
                event_tag_profile_feed):
    """Maps subfeeds to the corresponding ad.

    The Ad is an object that has several other dependent entities, they could be
    other entities like creative assignment, or complex sub objects in the ad
    entity like the placement assignment. This function maps those feeds by ID
    and injects the child feeds into the feed item representing the ad.

    Also, the ad level is where placement event tag profiles are assigned, and
    therefore this function is also responsible to determining if the placement
    event tag profile should be used, or if the direct event tag assignment in
    the ad should be used.

    Args:
      ad_feed: Ad feed.
      ad_creative_assignment: Ad creative assignment feed.
      ad_placement_assignment: Ad placement assignment feed.
      placement_feed: Placement feed.
      event_tag_profile_feed: Event tag profile feed.
    """
    for ad in ad_feed:
      ad['creative_assignment'] = [
          association for association in ad_creative_assignment
          if self._assignment_matches(ad, association)
      ]

      ad['placement_assignment'] = [
          association for association in ad_placement_assignment
          if self._assignment_matches(ad, association)
      ]

      if ad.get(FieldMap.PLACEMENT_ID, None) or ad.get(FieldMap.PLACEMENT_NAME,
                                                       None):
        ad['placement_assignment'].append(ad)

      ad['event_tag_assignment'] = [
          association for association in ad_event_tag_assignment
          if self._assignment_matches(ad, association)
      ]

      if ad.get(FieldMap.EVENT_TAG_ID, None) or ad.get(FieldMap.EVENT_TAG_NAME,
                                                       None):
        ad['event_tag_assignment'].append(ad)

      # Identify all event tag profiles associated with the placements
      ad['placement_event_tag_profile'] = []
      for placement_assignment in ad['placement_assignment']:
        placement = self._placement_dao.get(placement_assignment, required=True)

        if placement:
          ad_placement = None
          for item in placement_feed:
            if int(placement['id']) == item.get(FieldMap.PLACEMENT_ID, None):
              ad_placement = item

          if ad_placement:
            event_tag_profile_name = ad_placement.get(
                FieldMap.EVENT_TAG_PROFILE_NAME, '')

            if event_tag_profile_name:
              ad['placement_event_tag_profile'] += [
                  event_tag_profile
                  for event_tag_profile in event_tag_profile_feed
                  if event_tag_profile.get(FieldMap.EVENT_TAG_PROFILE_NAME,
                                           None) == event_tag_profile_name
              ]

  def _setup_rotation_strategy(self, creative_rotation, feed_item):
    """Analyzes the feed and sets up rotation strategy for the ad.

    For better user experience, the creative rotaion values that come from the
    feed map directly to values in the UI, this function is responsible for
    translating that to API specific values for the creative rotation objects
    under the ad.

    Args:
      creative_rotation: Feed item representing the creative rotation setup from
        the feed.
      feed_item: Feed item representing the ad.
    """
    option = feed_item.get(FieldMap.CREATIVE_ROTATION, 'Even').upper()

    if option == 'EVEN':
      creative_rotation['type'] = 'CREATIVE_ROTATION_TYPE_RANDOM'
      creative_rotation['weightCalculationStrategy'] = 'WEIGHT_STRATEGY_EQUAL'
    elif option == 'SEQUENTIAL':
      creative_rotation['type'] = 'CREATIVE_ROTATION_TYPE_SEQUENTIAL'
      creative_rotation['weightCalculationStrategy'] = None
    elif option == 'CUSTOM':
      creative_rotation['type'] = 'CREATIVE_ROTATION_TYPE_RANDOM'
      creative_rotation['weightCalculationStrategy'] = 'WEIGHT_STRATEGY_CUSTOM'
    elif option == 'CLICK-THROUGH RATE':
      creative_rotation['type'] = 'CREATIVE_ROTATION_TYPE_RANDOM'
      creative_rotation[
          'weightCalculationStrategy'] = 'WEIGHT_STRATEGY_HIGHEST_CTR'
    elif option == 'OPTIMIZED':
      creative_rotation['type'] = 'CREATIVE_ROTATION_TYPE_RANDOM'
      creative_rotation[
          'weightCalculationStrategy'] = 'WEIGHT_STRATEGY_OPTIMIZED'

  def _process_update(self, item, feed_item):
    """Updates an ad based on the values from the feed.

    Args:
      item: Object representing the ad to be updated, this object is updated
        directly.
      feed_item: Feed item representing ad values from the Bulkdozer feed.
    """
    campaign = self._campaign_dao.get(feed_item, required=True)

    item['active'] = feed_item.get(FieldMap.AD_ACTIVE, True)

    if item['active']:
      self._wait_all_creative_activation(feed_item)

    self._setup_rotation_strategy(item['creativeRotation'], feed_item)

    if feed_item['creative_assignment']:
      item['creativeRotation']['creativeAssignments'] = []

    item['placementAssignments'] = []
    item['eventTagOverrides'] = []

    self._process_assignments(
        feed_item, item['creativeRotation'].get('creativeAssignments', []),
        item['placementAssignments'], item['eventTagOverrides'], campaign)

    if 'deliverySchedule' in item:
      item['deliverySchedule']['priority'] = feed_item.get(
          FieldMap.AD_PRIORITY, None)

    if feed_item.get(FieldMap.AD_HARDCUTOFF, None) != None:
      if not 'deliverySchedule' in item:
        item['deliverySchedule'] = {}

      item['deliverySchedule']['hardCutoff'] = feed_item.get(
          FieldMap.AD_HARDCUTOFF)

    item['archived'] = feed_item.get(FieldMap.AD_ARCHIVED, False)

    if 'T' in feed_item.get(FieldMap.AD_END_DATE, None):
      item['endTime'] = feed_item.get(FieldMap.AD_END_DATE, None)
    else:
      item['endTime'] = StringExtensions.convertDateStrToDateTimeStr(
          feed_item.get(FieldMap.AD_END_DATE, None), '23:59:59')

    if 'T' in feed_item.get(FieldMap.AD_START_DATE, None):
      item['startTime'] = feed_item.get(FieldMap.AD_START_DATE, None)
    else:
      item['startTime'] = StringExtensions.convertDateStrToDateTimeStr(
          feed_item.get(FieldMap.AD_START_DATE, None))

    item['name'] = feed_item.get(FieldMap.AD_NAME, None)

    self._process_landing_page(item, feed_item)

  def _process_assignments(self, feed_item, creative_assignments,
                           placement_assignments, event_tag_assignments,
                           campaign):
    """Updates the ad by setting the values of child objects based on secondary feeds.

    Args:
      feed_item: Feed item representing the ad from the Bulkdozer feed.
      creative_assignments: Feed items representing creative assignments related
        with the current ad.
      placement_assignments: Feed items representing placement assignments
        related with the current ad.
      event_tag_assignments: Feed items representing event tag assignments
        related with the current ad.
    """
    assigned_creatives = []
    assigned_placements = []
    assigned_event_tags = []

    for assignment in feed_item['creative_assignment']:
      creative = self._creative_dao.get(assignment, required=True)
      assignment[FieldMap.CREATIVE_ID] = creative['id']

      if not creative['id'] in assigned_creatives:
        assigned_creatives.append(creative['id'])

        sequence = assignment.get(FieldMap.CREATIVE_ROTATION_SEQUENCE, None)
        weight = assignment.get(FieldMap.CREATIVE_ROTATION_WEIGHT, None)

        sequence = sequence if type(sequence) is int else None
        weight = weight if type(weight) is int else None

        if assignment.get(FieldMap.AD_CREATIVE_ROTATION_START_TIME, ''):
          startTime = (
              assignment.get(FieldMap.AD_CREATIVE_ROTATION_START_TIME,
                             '') if 'T' in assignment.get(
                                 FieldMap.AD_CREATIVE_ROTATION_START_TIME, '')
              else StringExtensions.convertDateStrToDateTimeStr(
                  feed_item.get(FieldMap.AD_CREATIVE_ROTATION_START_TIME,
                                None)))
          assignment[FieldMap.AD_CREATIVE_ROTATION_START_TIME] = startTime
        else:
          startTime = None

        if assignment.get(FieldMap.AD_CREATIVE_ROTATION_END_TIME, ''):
          endTime = (
              assignment.get(FieldMap.AD_CREATIVE_ROTATION_END_TIME, '') if
              'T' in assignment.get(FieldMap.AD_CREATIVE_ROTATION_END_TIME, '')
              else StringExtensions.convertDateStrToDateTimeStr(
                  feed_item.get(FieldMap.AD_CREATIVE_ROTATION_END_TIME, None),
                  '23:59:59'))
          assignment[FieldMap.AD_CREATIVE_ROTATION_END_TIME] = endTime
        else:
          endTime = None

        lp = None
        if assignment.get(FieldMap.AD_LANDING_PAGE_ID,
                          '') != 'CAMPAIGN_DEFAULT':
          lp = self._landing_page_dao.get(assignment, required=True)
        else:
          lp = self._landing_page_dao.get(
              {FieldMap.AD_LANDING_PAGE_ID: campaign['defaultLandingPageId']},
              required=True)

        creative_assignment = {
            'active': True,
            'sequence': sequence,
            'weight': weight,
            'creativeId': assignment.get(FieldMap.CREATIVE_ID, None),
            'startTime': startTime,
            'endTime': endTime,
            'clickThroughUrl': {
                'defaultLandingPage':
                    False if
                    (assignment.get(FieldMap.AD_LANDING_PAGE_ID, '') or
                     assignment.get(FieldMap.CUSTOM_CLICK_THROUGH_URL, '')) and
                    assignment.get(FieldMap.AD_LANDING_PAGE_ID,
                                   '') != 'CAMPAIGN_DEFAULT' else True,
                'landingPageId':
                    lp.get('id', None) if lp else None,
                'customClickThroughUrl':
                    assignment.get(FieldMap.CUSTOM_CLICK_THROUGH_URL, '')
            }
        }

        if creative.get('exitCustomEvents'):
          creative_assignment['richMediaExitOverrides'] = []

          if assignment.get(FieldMap.AD_LANDING_PAGE_ID, '') or assignment.get(
              FieldMap.CUSTOM_CLICK_THROUGH_URL, ''):
            for exit_custom_event in creative.get('exitCustomEvents', []):
              creative_assignment['richMediaExitOverrides'].append({
                  'exitId': exit_custom_event['id'],
                  'enabled': True,
                  'clickThroughUrl': {
                      'defaultLandingPage':
                          False if
                          (assignment.get(FieldMap.AD_LANDING_PAGE_ID, '') or
                           assignment.get(FieldMap.CUSTOM_CLICK_THROUGH_URL, '')
                          ) and assignment.get(FieldMap.AD_LANDING_PAGE_ID, '')
                          != 'CAMPAIGN_DEFAULT' else True,
                      'landingPageId':
                          lp.get('id', None) if lp else None,
                      'customClickThroughUrl':
                          assignment.get(FieldMap.CUSTOM_CLICK_THROUGH_URL, '')
                  }
              })

        creative_assignments.append(creative_assignment)

    for assignment in feed_item['placement_assignment']:
      placement = self._placement_dao.get(assignment, required=True)
      if placement:
        assignment[FieldMap.PLACEMENT_ID] = placement['id']

        if not placement['id'] in assigned_placements:
          assigned_placements.append(placement['id'])

          placement_assignments.append({
              'active': True,
              'placementId': assignment.get(FieldMap.PLACEMENT_ID, None),
          })

    event_tags = [{
        'assignment': item,
        'event_tag': self._event_tag_dao.get(item, required=True)
    } for item in feed_item['event_tag_assignment']]

    event_tags += [{
        'assignment': item,
        'event_tag': self._event_tag_dao.get(item, required=True)
    } for item in feed_item['placement_event_tag_profile']]

    for item in event_tags:
      assignment = item['assignment']
      event_tag = item['event_tag']

      if event_tag:
        assignment[FieldMap.EVENT_TAG_ID] = event_tag['id']

        if not event_tag['id'] in assigned_event_tags:
          assigned_event_tags.append(event_tag['id'])

          event_tag_assignments.append({
              'id': event_tag['id'],
              'enabled': assignment.get(FieldMap.EVENT_TAG_ENABLED, True)
          })

  def _process_new(self, feed_item):
    """Creates a new ad DCM object from a feed item representing an ad from the Bulkdozer feed.

    This function simply creates the object to be inserted later by the BaseDAO
    object.

    Args:
      feed_item: Feed item representing the ad from the Bulkdozer feed.

    Returns:
      An ad object ready to be inserted in DCM through the API.

    """
    if feed_item.get(FieldMap.AD_ACTIVE, None):
      self._wait_all_creative_activation(feed_item)

    campaign = self._campaign_dao.get(feed_item, required=True)

    creative_assignments = []
    placement_assignments = []
    event_tag_assignments = []
    self._process_assignments(feed_item, creative_assignments,
                              placement_assignments, event_tag_assignments,
                              campaign)

    creative_rotation = {'creativeAssignments': creative_assignments}

    self._setup_rotation_strategy(creative_rotation, feed_item)

    delivery_schedule = {
        'impressionRatio': '1',
        'priority': feed_item.get(FieldMap.AD_PRIORITY, None),
        'hardCutoff': feed_item.get(FieldMap.AD_HARDCUTOFF, None)
    }

    ad = {
        'active':
            feed_item.get(FieldMap.AD_ACTIVE, None),
        'archived':
            feed_item.get(FieldMap.AD_ARCHIVED, None),
        'campaignId':
            campaign['id'],
        'creativeRotation':
            creative_rotation,
        'deliverySchedule':
            delivery_schedule,
        'endTime':
            feed_item.get(FieldMap.AD_END_DATE, None) if 'T' in feed_item.get(
                FieldMap.AD_END_DATE, None) else
            StringExtensions.convertDateStrToDateTimeStr(
                feed_item.get(FieldMap.AD_END_DATE, None), '23:59:59'),
        'name':
            feed_item.get(FieldMap.AD_NAME, None),
        'placementAssignments':
            placement_assignments,
        'startTime':
            feed_item.get(FieldMap.AD_START_DATE, None) if 'T' in feed_item.get(
                FieldMap.AD_START_DATE, None) else
            StringExtensions.convertDateStrToDateTimeStr(
                feed_item.get(FieldMap.AD_START_DATE, None)),
        'type':
            feed_item.get(FieldMap.AD_TYPE, 'AD_SERVING_STANDARD_AD'),
        'eventTagOverrides':
            event_tag_assignments
    }

    self._process_landing_page(ad, feed_item)

    return ad

  def _process_landing_page(self, item, feed_item):
    """Configures ad landing page.

    Args:
      item: DCM ad object to update.
      feed_item: Feed item representing the ad from the Bulkdozer feed
    """
    if feed_item.get(FieldMap.AD_LANDING_PAGE_ID, ''):

      landing_page = self._landing_page_dao.get(feed_item, required=True)
      item['clickThroughUrl'] = {'landingPageId': landing_page['id']}

    if feed_item.get(FieldMap.AD_URL_SUFFIX, ''):
      item['clickThroughUrlSuffixProperties'] = {
          'overrideInheritedSuffix': True,
          'clickThroughUrlSuffix': feed_item.get(FieldMap.AD_URL_SUFFIX, '')
      }
    else:
      item['clickThroughUrlSuffixProperties'] = {
          'overrideInheritedSuffix': False
      }

  def _sub_entity_map(self, assignments, item, campaign):
    """Maps ids and names of sub entities so they can be updated in the Bulkdozer feed.

    When Bulkdozer is done processing an item, it writes back the updated names
    and ids of related objects, this method makes sure those are updated in the
    ad feed.

    Args:
      assignments: List of child feeds to map.
      item: The DCM ad object that was updated or created.
      campaign: The campaign object associated with the ad.
    """
    for assignment in assignments:
      placement = self._placement_dao.get(assignment, required=True)
      event_tag = self._event_tag_dao.get(assignment, required=True)
      creative = self._creative_dao.get(assignment, required=True)

      landing_page = None
      if assignment.get(FieldMap.AD_LANDING_PAGE_ID, '') != 'CAMPAIGN_DEFAULT':
        landing_page = self._landing_page_dao.get(assignment, required=True)

      if landing_page:
        assignment[FieldMap.AD_LANDING_PAGE_ID] = landing_page['id']

      if item:
        assignment[FieldMap.AD_ID] = item['id']
        assignment[FieldMap.AD_NAME] = item['name']

      if campaign:
        assignment[FieldMap.CAMPAIGN_ID] = campaign['id']
        assignment[FieldMap.CAMPAIGN_NAME] = campaign['name']

      if placement:
        assignment[FieldMap.PLACEMENT_ID] = placement['id']
        assignment[FieldMap.PLACEMENT_NAME] = placement['name']

      if creative:
        assignment[FieldMap.CREATIVE_ID] = creative['id']
        assignment[FieldMap.CREATIVE_NAME] = creative['name']

      if event_tag:
        assignment[FieldMap.EVENT_TAG_ID] = event_tag['id']
        assignment[FieldMap.EVENT_TAG_NAME] = event_tag['name']

  def _post_process(self, feed_item, item):
    """Maps ids and names of related entities so they can be updated in the Bulkdozer feed.

    When Bulkdozer is done processing an item, it writes back the updated names
    and ids of related objects, this method makes sure those are updated in the
    ad feed.

    Args:
      feed_item: Feed item representing the ad from the Bulkdozer feed.
      item: The DCM ad being updated or created.
    """
    campaign = self._campaign_dao.get(feed_item, required=True)
    feed_item[FieldMap.CAMPAIGN_ID] = campaign['id']
    feed_item[FieldMap.CAMPAIGN_NAME] = campaign['name']

    landing_page = self._landing_page_dao.get(feed_item, required=True)

    if landing_page:
      feed_item[FieldMap.AD_LANDING_PAGE_ID] = landing_page['id']

    self._sub_entity_map(feed_item['creative_assignment'], item, campaign)
    self._sub_entity_map(feed_item['placement_assignment'], item, campaign)
    self._sub_entity_map(feed_item['event_tag_assignment'], item, campaign)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Rich_Media_Schema = [{
    'name': 'Partitiontime',
    'type': 'TIMESTAMP',
    'mode': 'NULLABLE'
}, {
    'name': 'Event_Time',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'User_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Rendering_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Version',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Id_Dcm',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Country_Code',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'State_Region',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Browser_Platform_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Browser_Platform_Version',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Operating_System_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner_1_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner_2_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner_3_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner_4_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Partner_5_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Null_User_Id_Reason_Groups',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Event_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Event_Type_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Impression_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Pixel_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Custom_Event_Counters',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Custom_Event_Timers',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Event_Counters',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Event_Timers',
    'type': 'FLOAT',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from collections import OrderedDict

from starthinker.util.data import put_rows
from starthinker.util.google_api import API_Analytics
from starthinker.util.project import project
from starthinker.task.ga_settings_download.ga_schemas import CUSTOM_DIMENSION_SCHEMA
from starthinker.task.ga_settings_download.ga_schemas import CUSTOM_METRIC_SCHEMA
from starthinker.task.ga_settings_download.ga_schemas import VIEW_SCHEMA
from starthinker.task.ga_settings_download.ga_schemas import GOAL_SCHEMA
from starthinker.task.ga_settings_download.ga_schemas import GOOGLE_ADS_LINK_SCHEMA
from starthinker.task.ga_settings_download.ga_schemas import REMARKETING_AUDIENCE_SCHEMA
from starthinker.task.ga_settings_download.ga_schemas import ACCOUNT_SUMMARIES_SCHEMA


def custom_dimensions_download(accounts, current_date):
  if project.verbose:
    print('Downloading GA Custom Dimensions')

  for account in accounts:
    for web_property in account.get('webProperties', []):
      for custom_dimension in API_Analytics(
          project.task['auth'],
          iterate=True).management().customDimensions().list(
              accountId=account.get('id'),
              webPropertyId=web_property.get('id')).execute():
        yield (account.get('name'), account.get('id'), web_property.get('name'),
               web_property.get('id'), custom_dimension.get('id'),
               custom_dimension.get('name'), custom_dimension.get('index'),
               custom_dimension.get('scope'), custom_dimension.get('active'),
               custom_dimension.get('created'), custom_dimension.get('updated'),
               current_date, custom_dimension.get('selfLink'))


def custom_metrics_download(accounts, current_date):
  if project.verbose:
    print('Downloading GA Custom Metrics')

  for account in accounts:
    for web_property in account.get('webProperties', []):
      for custom_metric in API_Analytics(
          project.task['auth'], iterate=True).management().customMetrics().list(
              accountId=account.get('id'),
              webPropertyId=web_property.get('id')).execute():
        yield (account.get('name'), account.get('id'), web_property.get('name'),
               web_property.get('id'), custom_metric.get('id'),
               custom_metric.get('name'), custom_metric.get('index'),
               custom_metric.get('scope'), custom_metric.get('active'),
               custom_metric.get('created'), custom_metric.get('updated'),
               current_date, custom_metric.get('selfLink'),
               custom_metric.get('type'), custom_metric.get('min_value'),
               custom_metric.get('max_value'))


def goals_download(accounts, current_date):
  if project.verbose:
    print('Downloading GA Goals')

  for goal in API_Analytics(
      project.task['auth'], iterate=True).management().goals().list(
          accountId='~all', webPropertyId='~all', profileId='~all').execute():
    for account in accounts:
      if goal.get('accountId') == account.get('id'):
        for prop in account.get('webProperties', []):
          if goal.get('webPropertyId') == prop.get('id'):
            for view in prop.get('profiles', []):
              if goal.get('profileId') == view.get('id'):
                ordered_goal = OrderedDict()
                ordered_goal['date'] = str(current_date)
                ordered_goal['id'] = goal.get('id')
                ordered_goal['accountId'] = goal.get('accountId')
                ordered_goal['webPropertyId'] = goal.get('webPropertyId')
                ordered_goal['profileId'] = goal.get('profileId')
                ordered_goal['internalWebPropertyId'] = goal.get(
                    'internalWebPropertyId')
                ordered_goal['name'] = goal.get('name')
                ordered_goal['accountName'] = account.get('name')
                ordered_goal['webPropertyName'] = prop.get('name')
                ordered_goal['profileName'] = view.get('name')
                ordered_goal['value'] = goal.get('value')
                ordered_goal['active'] = goal.get('active')
                ordered_goal['type'] = goal.get('type')
                ordered_goal['created'] = goal.get('created')
                ordered_goal['updated'] = goal.get('updated')
                if goal.get('urlDestinationDetails'):
                  ordered_goal['urlDestinationDetails'] = {
                      'url':
                          goal.get('urlDestinationDetails').get('url'),
                      'caseSensitive':
                          goal.get('urlDestinationDetails').get('caseSensitive'
                                                               ),
                      'matchType':
                          goal.get('urlDestinationDetails').get('matchType'),
                      'firstStepRequired':
                          goal.get('urlDestinationDetails').get(
                              'firstStepRequired')
                  }
                  if goal.get('urlDestinationDetails').get('steps', []):
                    ordered_goal['urlDestinationDetails']['steps'] = []
                    for step in goal.get('urlDestinationDetails').get(
                        'steps', []):
                      ordered_goal['urlDestinationDetails']['steps'].append({
                          'number': step.get('number'),
                          'name': step.get('name'),
                          'url': step.get('url')
                      })
                if goal.get('visitTimeOnSiteDetails'):
                  ordered_goal['visitTimeOnSiteDetails'] = {
                      'comparisonType':
                          goal.get('visitTimeOnSiteDetails').get(
                              'comparisonType'),
                      'comparisonValue':
                          goal.get('visitTimeOnSiteDetails').get(
                              'comparisonValue')
                  }
                if goal.get('visitNumPagesDetails'):
                  ordered_goal['visitNumPagesDetails'] = {
                      'comparisonType':
                          goal.get('visitNumPagesDetails').get('comparisonType'
                                                              ),
                      'comparisonValue':
                          goal.get('visitNumPagesDetails').get('comparisonValue'
                                                              )
                  }
                if goal.get('eventDetails'):
                  ordered_goal['eventDetails'] = {
                      'useEventValue':
                          goal.get('eventDetails').get('useEventValue'),
                      'eventConditions': []
                  }
                  for conditions in goal.get('eventDetails').get(
                      'eventConditions'):
                    ordered_goal['eventDetails']['eventConditions'].append({
                        'type': conditions.get('type'),
                        'matchType': conditions.get('matchType'),
                        'expression': conditions.get('expression'),
                        'comprisonType': conditions.get('comprisonType'),
                        'comparisonValue': conditions.get('comparisonValue')
                    })
                yield ordered_goal


def views_download(accounts, current_date):
  if project.verbose:
    print('Downloading GA views')

  for view in API_Analytics(
      project.task['auth'], iterate=True).management().profiles().list(
          accountId='~all', webPropertyId='~all').execute():
    for account in accounts:
      if view.get('accountId') == account.get('id'):
        for prop in account.get('webProperties', []):
          if view.get('webPropertyId') == prop.get('id'):
            yield (current_date, view.get('id'), view.get('selfLink'),
                   view.get('accountId'), view.get('webPropertyId'),
                   account.get('name'), prop.get('name'), view.get('name'),
                   view.get('currency'), view.get('timezone'),
                   view.get('websiteUrl'), view.get('defaultPage'),
                   view.get('excludeQueryParameters'),
                   view.get('siteSearchQueryParameters'),
                   view.get('stripSiteSearchQueryParameters'),
                   view.get('siteSearchCategoryParameters'),
                   view.get('stripSiteSearchCategoryParameters'),
                   view.get('type'), view.get('created'), view.get('updated'),
                   view.get('eCommerceTracking'),
                   view.get('enhancedECommerceTracking'),
                   view.get('botFilteringEnabled'), view.get('starred'))


def google_ads_links_download(accounts, current_date):
  if project.verbose:
    print('Downloading GA Google Ads Links')

  for account in accounts:
    for prop in account.get('webProperties', []):
      for link in API_Analytics(
          project.task['auth'],
          iterate=True).management().webPropertyAdWordsLinks().list(
              accountId=account.get('id'),
              webPropertyId=prop.get('id')).execute():
        ordered_link = OrderedDict()
        ordered_link['date'] = str(current_date)
        ordered_link['id'] = link.get('id')
        ordered_link['kind'] = link.get('kind')
        ordered_link['selfLink'] = link.get('selfLink')
        ordered_link['entity'] = {
            'webPropertyRef': {
                'id':
                    link.get('entity').get('webPropertyRef').get('id'),
                'kind':
                    link.get('entity').get('webPropertyRef').get('kind'),
                'href':
                    link.get('entity').get('webPropertyRef').get('href'),
                'accountId':
                    link.get('entity').get('webPropertyRef').get('accountId'),
                'internalWebPropertyId':
                    link.get('entity').get('webPropertyRef').get(
                        'internalWebPropertyId'),
                'name':
                    link.get('entity').get('webPropertyRef').get('name')
            }
        }
        ordered_link['adWordsAccounts'] = []
        for ad_account in link.get('adWordsAccounts'):
          ordered_link['adWordsAccounts'].append({
              'kind': ad_account.get('kind'),
              'customerId': ad_account.get('customerId'),
              'autoTaggingEnabled': ad_account.get('autoTaggingEnabled')
          })
        ordered_link['name'] = link.get('name')
        ordered_link['profileIds'] = []
        for profile_id in link.get('profileIds', []):
          ordered_link['profileIds'].append({'id': profile_id})
        yield ordered_link


def remarketing_audiences_download(accounts, current_date):
  if project.verbose:
    print('Downloading GA Google Remarketing Audiences')

  for account in accounts:
    for prop in account.get('webProperties', []):
      for audience in API_Analytics(
          project.task['auth'],
          iterate=True).management().remarketingAudience().list(
              accountId=account.get('id'),
              webPropertyId=prop.get('id')).execute():
        ordered_audience = OrderedDict()
        ordered_audience['date'] = str(current_date)
        ordered_audience['kind'] = audience.get('kind')
        ordered_audience['id'] = audience.get('id')
        ordered_audience['accountId'] = audience.get('accountId')
        ordered_audience['webPropertyId'] = audience.get('webPropertyId')
        ordered_audience['webPropertyName'] = audience.get('name')
        ordered_audience['accountName'] = audience.get('name')
        ordered_audience['internalWebPropertyId'] = audience.get(
            'internalWebPropertyId')
        ordered_audience['created'] = audience.get('created')
        ordered_audience['updated'] = audience.get('updated')
        ordered_audience['name'] = audience.get('name')
        ordered_audience['description'] = audience.get('description')
        ordered_audience['linkedAdAccounts'] = []
        for linked_ad_account in audience.get('linkedAdAccounts', []):
          ordered_audience['linkedAdAccounts'].append({
              'kind':
                  linked_ad_account.get('kind'),
              'id':
                  linked_ad_account.get('id'),
              'accountId':
                  linked_ad_account.get('accountId'),
              'webPropertyId':
                  linked_ad_account.get('webPropertyId'),
              'internalWebPropertyId':
                  linked_ad_account.get('internalWebPropertyId'),
              'remarketingAudienceId':
                  linked_ad_account.get('remarketingAudienceId'),
              'linkedAccountId':
                  linked_ad_account.get('linkedAccountId'),
              'type':
                  linked_ad_account.get('type'),
              'status':
                  linked_ad_account.get('status'),
              'eligibleForSearch':
                  linked_ad_account.get('eligibleForSearch')
          })
        ordered_audience['linkedViews'] = []
        for linked_view in audience.get('linkedViews', []):
          ordered_audience['linkedViews'].append({'id': linked_view})
        ordered_audience['audienceType'] = audience.get('audienceType')
        if audience.get('audienceType') == 'SIMPLE':
          ordered_audience['audienceDefinition'] = {
              'includeConditions': {
                  'kind':
                      audience.get('audienceDefinition').get('includeConditions'
                                                            ).get('kind'),
                  'isSmartList':
                      audience.get('audienceDefinition').get(
                          'includeConditions').get('isSmartList'),
                  'segment':
                      audience.get('audienceDefinition').get('includeConditions'
                                                            ).get('segment'),
                  'membershipDurationDays':
                      audience.get('audienceDefinition').get(
                          'includeConditions').get('membershipDurationDays'),
                  'daysToLookBack':
                      audience.get('audienceDefinition').get(
                          'includeConditions').get('daysToLookBack')
              }
          }
        elif audience.get('audienceType') == 'STATE_BASED':
          if audience.get('stateBasedAudienceDefinition').get(
              'includeConditions'):
            ordered_audience['stateBasedAudienceDefinition'] = {
                'includeConditions': {
                    'kind':
                        audience.get('stateBasedAudienceDefinition').get(
                            'includeConditions').get('kind'),
                    'isSmartList':
                        audience.get('stateBasedAudienceDefinition').get(
                            'includeConditions').get('isSmartList'),
                    'segment':
                        audience.get('stateBasedAudienceDefinition').get(
                            'includeConditions').get('segment'),
                    'membershipDurationDays':
                        audience.get('stateBasedAudienceDefinition').get(
                            'includeConditions').get('membershipDurationDays'),
                    'daysToLookBack':
                        audience.get('stateBasedAudienceDefinition').get(
                            'includeConditions').get('daysToLookBack')
                }
            }
          if audience.get('stateBasedAudienceDefinition').get(
              'excludeConditions'):
            ordered_audience['stateBasedAudienceDefinition'] = {
                'excludeConditions': {
                    'segment':
                        audience.get('stateBasedAudienceDefinition').get(
                            'excludeConditions').get('segment'),
                    'exclusionDuration':
                        audience.get('stateBasedAudienceDefinition').get(
                            'excludeConditions').get('exclusionDuration')
                }
            }
        yield ordered_audience


def account_summaries_download(accounts, current_date):
  if project.verbose:
    print('Downloading GA Google Remarketing Audiences')

  for account in accounts:
    ordered_summary = OrderedDict()
    ordered_summary['date'] = str(current_date)
    ordered_summary['id'] = account.get('id')
    ordered_summary['kind'] = account.get('kind')
    ordered_summary['name'] = account.get('name')
    ordered_summary['starred'] = account.get('starred')
    ordered_summary['webProperties'] = []
    for prop in account.get('webProperties', []):
      ordered_prop = OrderedDict()
      ordered_prop['kind'] = prop.get('kind')
      ordered_prop['id'] = prop.get('id')
      ordered_prop['name'] = prop.get('name')
      ordered_prop['internalWebPropertyId'] = prop.get('internalWebPropertyId')
      ordered_prop['level'] = prop.get('level')
      ordered_prop['websiteUrl'] = prop.get('websiteUrl')
      ordered_prop['starred'] = prop.get('starred')
      ordered_prop['profiles'] = []
      for profile in prop.get('profiles', []):
        ordered_prop['profiles'].append({
            'kind': profile.get('kind'),
            'id': profile.get('id'),
            'name': profile.get('name'),
            'type': profile.get('type'),
            'starred': profile.get('starred'),
        })
      ordered_summary['webProperties'].append(ordered_prop)
    yield ordered_summary


def write_to_bigquery(table_id, schema, data, data_format):
  out = {
      'bigquery': {
          'format': data_format,
          'dataset': project.task['dataset'],
          'table': table_id,
          'schema': schema,
          'skip_rows': 0,
          'disposition': 'WRITE_TRUNCATE'
      }
  }
  put_rows(project.task['auth'], out, data)


@project.from_parameters
def ga_settings_download():
  if project.verbose:
    print('Initiating GA Settings Download')

  filters = [str(f) for f in project.task.get('accounts', [])
            ]  # in case integer list is provided
  accounts = [
      a for a in API_Analytics(project.task['auth'], iterate=True).management()
      .accountSummaries().list().execute()
      if not filters or a['id'] in filters or a['name'] in filters
  ]

  if accounts:
    current_date = project.date

    write_to_bigquery('ga_custom_dimension_settings', CUSTOM_DIMENSION_SCHEMA,
                      custom_dimensions_download(accounts, current_date), 'CSV')

    write_to_bigquery('ga_custom_metric_settings', CUSTOM_METRIC_SCHEMA,
                      custom_metrics_download(accounts, current_date), 'CSV')

    write_to_bigquery('ga_view_settings', VIEW_SCHEMA,
                      views_download(accounts, current_date), 'CSV')

    write_to_bigquery('ga_goal_settings', GOAL_SCHEMA,
                      goals_download(accounts, current_date), 'JSON')

    write_to_bigquery('ga_google_ad_link_settings', GOOGLE_ADS_LINK_SCHEMA,
                      google_ads_links_download(accounts, current_date), 'JSON')

    write_to_bigquery('ga_remarketing_audience_settings',
                      REMARKETING_AUDIENCE_SCHEMA,
                      remarketing_audiences_download(accounts,
                                                     current_date), 'JSON')

    write_to_bigquery('ga_account_summaries_settings', ACCOUNT_SUMMARIES_SCHEMA,
                      account_summaries_download(accounts, current_date),
                      'JSON')


if __name__ == '__main__':
  ga_settings_download()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re

from starthinker.util.project import get_project

RE_TEXT_FIELD = re.compile(r'\{(.*?:.*?)\}')


def fields_to_string(fields, values={}):
  items = [
      repr(field['name']) + ': ' +
      repr(values.get(field['name'], field.get('default', ''))) + ',' +
      ('  # %s' % field['description'] if 'description' in field else '')
      for field in fields
  ]
  return '{\n  %s\n}' % ('\n  '.join(items))


def dict_to_string(value, char_indent='  ', char_line='\n', skip=[], indent=0):
  nlch = char_line + char_indent * (indent + 1)
  if type(value) is dict:
    is_skip = any(k in value for k in skip)
    items = [('' if is_skip else nlch) + repr(key) + ': ' +
             dict_to_string(value[key], '' if is_skip else char_indent,
                            '' if is_skip else char_line, skip, indent + 1)
             for key in value]
    return '{%s}' % (','.join(items) +
                     ('' if is_skip else char_line + char_indent * indent))
  elif type(value) is list:
    items = [
        nlch + dict_to_string(item, char_indent, char_line, skip, indent + 1)
        for item in value
    ]
    return '[%s]' % (','.join(items) + char_line + char_indent * indent)
  elif type(value) is tuple:
    items = [
        nlch + dict_to_string(item, char_indent, char_line, skip, indent + 1)
        for item in value
    ]
    return '(%s)' % (','.join(items) + char_line + char_indent * indent)
  else:
    return repr(value)


def json_set_auths(struct, auth):
  """Recusrsively finds auth in script JSON and sets them.

    Args:
      struct: (dict) A dictionary representation fo the JSON script.
      auth: (string) Either 'service' or 'user'.

    Returns:
      (struct) same structure but with all auth fields replaced.

  """

  if isinstance(struct, dict):
    if 'auth' in struct:
      struct['auth'] = auth
    for key, value in struct.items():
      json_set_auths(value, auth)
  elif isinstance(struct, list) or isinstance(struct, tuple):
    for index, value in enumerate(struct):
      json_set_auths(value, auth)

  return struct


def json_get_fields(struct, path=[]):
  """Recusrsively finds fields in script JSON and returns them as a list.

     Field has format: { "field":{ "name":"???", "kind":"???", "default":???,
     "description":"???" }}

    Args:
      struct: (dict) A dictionary representation fo the JSON script.
      path: (list) Stack that keeps track of recursion depth. Not used
        externally.

    Returns:
      fields: (list) A list of dictionaries representing each field struct found
      in the JSON.

  """

  fields = {}
  path = path[:]
  if isinstance(struct, dict):
    if 'field' in struct:
      fields[struct['field']['name']] = struct['field']
    else:
      for key, value in struct.items():
        fields.update(json_get_fields(value, path + [key]))
  elif isinstance(struct, list) or isinstance(struct, tuple):
    for index, value in enumerate(struct):
      fields.update(json_get_fields(value, path + [index]))

  if path == []:
    return sorted(
        fields.values(),
        key=lambda f: f.get('order', 0))  # sort only on last step of recursion
  else:
    return fields  # do not sort if deep in recursion


def get_field_value(field, variables):
  value = None
  try:
    value = variables.get(field['name'], field.get('default'))
    if value is not None and 'prefix' in field:
      value = '%s%s' % (field['prefix'], value)
  except KeyError:
    pass

  return value


def json_set_fields(struct, variables):
  """Recusrsively replaces fields in script JSON with values provided.

     Field has format: { "field":{ "name":"???", "kind":"???", "default":???,
     "description":"???" }}

     If field value is empty and field default is null, the value is removed
     from JSON as a parameter,
     allowing the python task to pick a default value. Allows optional
     parameters to exist.

    Args:
      struct: (dict) A dictionary representation of the JSON script.
      variables: (dict) A lookup table of all values to be replaced, key is name
        of field.

    Returns:
      Nothig. Struct is modified in place.

  """

  if isinstance(struct, dict):
    for key, value in list(struct.items()):
      if isinstance(value, dict) and 'field' in value:
        variable_value = get_field_value(value['field'], variables)
        if variable_value is None and value.get('default') is None:
          del struct[key]
        else:
          struct[key] = get_field_value(value['field'], variables)
      else:
        json_set_fields(value, variables)
  elif isinstance(struct, list) or isinstance(struct, tuple):
    for index, value in enumerate(struct):
      if isinstance(value, dict) and 'field' in value:
        struct[index] = get_field_value(value['field'], variables)
      else:
        json_set_fields(value, variables)


def text_set_fields(text, variables):
  """Replaces fields in text with values from recipe.

     Fields are {field:[string]} or {field:[string], prefix:[string]} where
     field is a key in variables
     and prefix is a string value that gets appended to the value from
     variables.

     Args:
       text (string) A paragraph containing {field:[string]} or {field:[string],
         prefix:[string]}.
       variables: (dict) The keys mapping to field, and values to replace those
         fields.

     Returns:
       A string with all the {field:[string]} or {field:[string],
       prefix:[string]} values replaced by actual values from variables.

  """

  for field in RE_TEXT_FIELD.findall(text):
    parts = dict([p.strip().split(':') for p in field.split(',', 1)])
    value = (parts.get('prefix', '') + variables[parts['field']]
            ) if parts.get('field') in variables else 'UNDEFINED'
    text = text.replace('{' + field + '}', value)
  return text


def json_set_instructions(struct, variables):
  """Replaces all fields in instructions with values provided.

     Checks if struct['script']['instructions'] exist.  The replaces all %(???)s
     variables
     with values provided.  Note: %(???)s must match { "field":{ "name":"???" }}
     in JOSN.

    Args:
      struct: (dict) A dictionary representation of the JSON script.
      variables: (dict) A lookup table of all values to be replaced, key is name
        of field.

    Returns:
      Nothig. Instructions are modified in place.

  """

  if 'script' in struct:
    if 'instructions' in struct['script']:
      try:
        struct['script']['instructions'] = [
            text_set_fields(instruction, variables)
            for instruction in struct['script']['instructions']
        ]
      except KeyError:
        pass


def json_set_description(struct, variables):
  """Replaces all fields in description with values provided.

     Checks if struct['script']['description'] exist.  The replaces all %(???)s
     variables
     with values provided.  Note: %(???)s must match { "field":{ "name":"???" }}
     in JOSN.

    Args:
      struct: (dict) A dictionary representation of the JSON script.
      variables: (dict) A lookup table of all values to be replaced, key is name
        of field.

    Returns:
      Nothig. Description is modified in place.

  """

  if 'script' in struct:
    if 'description' in struct['script']:
      try:
        struct['script']['description'] = text_set_fields(
            struct['script']['description'], variables)
      except KeyError:
        pass



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
  Fields to pull from Census and use in all transforms.

  Commented out fields DO NOT meet Google Personalized Advertising Policy.
  See: https://support.google.com/adspolicy/answer/143465?hl=en
'''

CENSUS_FIELDS = [{
    'category': None,
    'denominator': None,
    'columns': ['geo_id', 'total_pop']
}, {
    'category': 'gender',
    'denominator': 'total_pop',
    'columns': [
        'female_pop',
        'male_pop',
    ]
}, {
    'category':
        'age',
    'denominator':
        'total_pop',
    'columns': [
#        'population_1_year_and_over',
#        'population_3_years_over',
#        'pop_5_years_over',
#        'pop_15_and_over',
#        'pop_16_over',
        'pop_25_years_over',
        'pop_25_64',
    ]
}, {
    'category':
        'male age',
    'denominator':
        'total_pop',
    'columns': [
#        'male_5_to_9',
#        'male_10_to_14',
#        'male_15_to_17',
        'male_18_to_19',
        'male_20',
        'male_21',
        'male_22_to_24',
        'male_25_to_29',
        'male_30_to_34',
        'male_35_to_39',
        'male_40_to_44',
        'male_45_to_49',
        'male_50_to_54',
        'male_55_to_59',
        'male_65_to_66',
        'male_67_to_69',
        'male_70_to_74',
        'male_75_to_79',
        'male_80_to_84',
        'male_85_and_over',
    ]
}, {
    'category':
        'female age',
    'denominator':
        'total_pop',
    'columns': [
        #'female_5_to_9',
        #'female_10_to_14',
        #'female_15_to_17',
        'female_18_to_19',
        'female_20',
        'female_21',
        'female_22_to_24',
        'female_25_to_29',
        'female_30_to_34',
        'female_35_to_39',
        'female_40_to_44',
        'female_45_to_49',
        'female_50_to_54',
        'female_55_to_59',
        'female_65_to_66',
        'female_67_to_69',
        'female_70_to_74',
        'female_75_to_79',
        'female_80_to_84',
        'female_85_and_over',
    ]
#}, {
#    'category':
#        'race',
#    'denominator':
#        'total_pop',
#    'columns': [
#        'white_pop',
#        'black_pop',
#        'asian_pop',
#        'hispanic_pop',
#        'amerindian_pop',
#        'other_race_pop',
#        'two_or_more_races_pop',
#        'hispanic_any_race',
#        'not_hispanic_pop',
#    ]
#}, {
#    'category':
#        'race age',
#    'denominator':
#        'total_pop',
#    'columns': [
#        'asian_male_45_54',
#        'asian_male_55_64',
#        'black_male_45_54',
#        'black_male_55_64',
#        'hispanic_male_45_54',
#        'hispanic_male_55_64',
#        'white_male_45_54',
#        'white_male_55_64',
#    ]
}, {
    'category':
        'income',
    'denominator':
        'total_pop',
    'columns': [
        'income_less_10000',
        'income_10000_14999',
        'income_15000_19999',
        'income_20000_24999',
        'income_25000_29999',
        'income_30000_34999',
        'income_35000_39999',
        'income_40000_44999',
        'income_45000_49999',
        'income_50000_59999',
        'income_60000_74999',
        'income_75000_99999',
        'income_100000_124999',
        'income_125000_149999',
        'income_150000_199999',
        'income_200000_or_more',
    ]
#}, {
#    'category': 'poverty',
#    'denominator': None,
#    'columns': [
#        'poverty',
#        'gini_index',
#    ]
}, {
    'category':
        'housing type',
    'denominator':
        'housing_units',
    'columns': [
        'occupied_housing_units',
        'housing_units_renter_occupied',
        'vacant_housing_units',
        'vacant_housing_units_for_rent',
        'vacant_housing_units_for_sale',
        'owner_occupied_housing_units',
        'million_dollar_housing_units',
        'mortgaged_housing_units',
    ]
}, {
    'category':
        'housing size',
    'denominator':
        'housing_units',
    'columns': [
        'dwellings_1_units_detached',
        'dwellings_1_units_attached',
        'dwellings_2_units',
        'dwellings_3_to_4_units',
        'dwellings_5_to_9_units',
        'dwellings_10_to_19_units',
        'dwellings_20_to_49_units',
        'dwellings_50_or_more_units',
        'mobile_homes',
        'group_quarters',
    ]
}, {
    'category':
        'housing age',
    'denominator':
        'housing_units',
    'columns': [
        'housing_built_2005_or_later',
        'housing_built_2000_to_2004',
        'housing_built_1939_or_earlier',
    ]
}, {
    'category':
        'housing rent',
    'denominator':
        'housing_units',
    'columns': [
        'rent_over_50_percent',
        'rent_40_to_50_percent',
        'rent_35_to_40_percent',
        'rent_30_to_35_percent',
        'rent_25_to_30_percent',
        'rent_20_to_25_percent',
        'rent_15_to_20_percent',
        'rent_10_to_15_percent',
        'rent_under_10_percent',
        'rent_burden_not_computed',
    ]
}, {
    'category': 'housing rent',
    'denominator': 'housing_units',
    'columns': ['percent_income_spent_on_rent',]
#}, {
#    'category':
#        'housing mobility',
#    'denominator':
#        'housing_units',
#    'columns': [
#        'different_house_year_ago_different_city',
#        'different_house_year_ago_same_city'
#    ]
}, {
    'category':
        'family',
    'denominator':
        'households',
    'columns': [
        'married_households',
        'nonfamily_households',
        'family_households',
#        'households_public_asst_or_food_stamps',
        'male_male_households',
        'female_female_households',
        'children',
#        'children_in_single_female_hh',
        'families_with_young_children',
        'two_parent_families_with_young_children',
        'two_parents_in_labor_force_families_with_young_children',
        'two_parents_father_in_labor_force_families_with_young_children',
        'two_parents_mother_in_labor_force_families_with_young_children',
        'two_parents_not_in_labor_force_families_with_young_children',
#        'one_parent_families_with_young_children',
#        'father_one_parent_families_with_young_children',
#        'father_in_labor_force_one_parent_families_with_young_children',
    ]
}, {
    'category':
        'cars owned',
    'denominator':
        'total_pop',
    'columns': [
        'no_car',
        'one_car',
        'two_cars',
        'three_cars',
        'four_more_cars',
    ]
}, {
    'category':
        'commuter time',
    'denominator':
        'total_pop',
    'columns': [
        'commute_less_10_mins',
        'commute_10_14_mins',
        'commute_15_19_mins',
        'commute_20_24_mins',
        'commute_25_29_mins',
        'commute_30_34_mins',
        'commute_35_44_mins',
        'commute_45_59_mins',
    ]
}, {
    'category': 'commuter age',
    'denominator': 'total_pop',
    'columns': ['commuters_16_over',]
}, {
    'category':
        'commuter method',
    'denominator':
        'total_pop',
    'columns': [
        'worked_at_home',
        'walked_to_work',
        'commuters_by_public_transportation',
        'commuters_by_bus',
        'commuters_by_car_truck_van',
        'commuters_by_carpool',
        'commuters_by_subway_or_elevated',
        'commuters_drove_alone',
    ]
}, {
    'category':
        'education degree',
    'denominator':
        'total_pop',
    'columns': [
        'associates_degree',
        'bachelors_degree',
        'bachelors_degree_2',
        'bachelors_degree_or_higher_25_64',
        'high_school_diploma',
        'high_school_including_ged',
        'less_one_year_college',
        'masters_degree',
        'one_year_more_college',
        'less_than_high_school_graduate',
        'graduate_professional_degree',
        'some_college_and_associates_degree',
    ]
#}, {
#    'category':
#        'education grade',
#    'denominator':
#        'total_pop',
#    'columns': [
#        'in_school',
#        'in_grades_1_to_4',
#        'in_grades_5_to_8',
#        'in_grades_9_to_12',
#        'in_undergrad_college',
#    ]
}, {
    'category':
        'education male age',
    'denominator':
        'total_pop',
    'columns': [
        'male_45_64_associates_degree',
        'male_45_64_bachelors_degree',
        'male_45_64_graduate_degree',
        'male_45_64_less_than_9_grade',
        'male_45_64_grade_9_12',
        'male_45_64_high_school',
        'male_45_64_some_college',
    ]
}, {
    'category':
        'labor status',
    'denominator':
        'total_pop',
    'columns': [
        'employed_pop',
#        'unemployed_pop',
        'pop_in_labor_force',
        'not_in_labor_force',
        'civilian_labor_force',
        'armed_forces',
        'workers_16_and_over',
    ]
}, {
    'category':
        'labor occupation',
    'denominator':
        'total_pop',
    'columns': [
        'employed_agriculture_forestry_fishing_hunting_mining',
        'employed_arts_entertainment_recreation_accommodation_food',
        'employed_construction',
        'employed_education_health_social',
        'employed_finance_insurance_real_estate',
        'employed_information',
        'employed_manufacturing',
        'employed_other_services_not_public_admin',
        'employed_public_administration',
        'employed_retail_trade',
        'employed_science_management_admin_waste',
        'employed_transportation_warehousing_utilities',
        'employed_wholesale_trade',
        'occupation_management_arts',
        'occupation_natural_resources_construction_maintenance',
        'occupation_production_transportation_material',
        'occupation_sales_office',
        'occupation_services',
        'management_business_sci_arts_employed',
        'sales_office_employed',
    ]
#}, {
#    'category':
#        'language',
#    'denominator':
#        'total_pop',
#    'columns': [
#        'speak_only_english_at_home',
#        'speak_spanish_at_home',
#        'speak_spanish_at_home_low_english',
#    ]
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

class Queries:
  browser_env_90 = """
    SELECT
      Partner,
      Advertiser,
      Advertiser_Id,
      Insertion_Order,
      Insertion_Order_Id,
      Campaign,
      Campaign_Id,
      CASE
        when Line_Item_Type = "TrueView" THEN "TrueView"
        when Line_Item_Type != "TrueView" and Browser_Detail = "Other" THEN "Other"
        ELSE Browser
      END AS Browser,
      Environment,
      Device,
      CASE
        when Line_Item_Type = "TrueView" THEN "TrueView"
        ELSE Concat(Device,"-",Environment)
      END AS Device_Environment,
      Line_item_type,
      Week_start,
      SUM(Impressions) AS Impressions
    FROM
      `{dataset}.DV360_Browser_Report_Clean`
    WHERE
      DATE_DIFF(CURRENT_DATE(), Week_start, WEEK) < 12
    GROUP BY
      Partner,
      Advertiser,
      Advertiser_Id,
      Insertion_Order,
      Insertion_Order_Id,
      Campaign,
      Campaign_Id,
      Browser,
      Environment,
      Device,
      Device_Environment,
      Week_start,
      Line_item_type
  """

  browser_2_year = """
    SELECT
      Partner,
      Advertiser,
      Advertiser_ID,
      Campaign,
      Insertion_Order,
      CASE
        when Line_Item_Type = "TrueView" THEN "TrueView"
        when Line_Item_Type != "TrueView" and Browser_Detail = "Other" THEN "Other"
        ELSE Browser
      END AS Browser,
      Browser_detail as Browser_Detail,
      Environment,
      Device,
      CASE
        when Line_Item_Type = "TrueView" THEN "TrueView"
        ELSE Concat(Device,"-",Environment)
      END AS Device_Environment,
      Week,
      Month,
      Year,
      Week_start,
      Line_Item_Type,
      Segment1,
      Segment2,
      Segment3,
      SUM(Impressions) AS Impressions,
      SUM(Total_Conversions) AS Total_Conversions,
      SUM(Post_Click_Conversions) AS Post_Click_Conversions,
      SUM(Post_View_Conversions) AS Post_View_Conversions,
      SUM(Revenue_Adv_Currency) AS Revenue_Adv_Currency,
      SUM(Media_Cost_Advertiser_Currency) AS Media_Cost_Advertiser_Currency,
      SUM(CM_Post_View_Revenue) AS CM_Post_View_Revenue,
      SUM(CM_Post_Click_Revenue) AS CM_Post_Click_Revenue
    FROM
      `{dataset}.z_DV360_Browser_Report_Clean`
    GROUP BY
      Partner,
      Advertiser,
      Advertiser_ID,
      Campaign,
      Insertion_Order,
      Browser,
      Browser_detail,
      Environment,
      Device,
      Device_Environment,
      Week,
      Month,
      Year,
      Line_Item_Type,
      Segment1,
      Segment2,
      Segment3,
      Week_start
  """

  browser_share_multichart = """
    WITH
    filtered AS
    (
      SELECT
        segment1,
        segment2,
        segment3,
        browser,
        sum(Impressions) AS imps,
        IF(UPPER(browser)='SAFARI', SUM(Impressions), 0) AS safari_imps,
        Line_Item_Type,
        Advertiser,
        Advertiser_ID,
        Campaign,
        Insertion_Order,
        Device_Environment,
        Week_start
      FROM `{dataset}.DV3_Browser`
      GROUP BY
        segment1,
        segment2,
        segment3,
        browser,
        Line_Item_Type,
        Advertiser,
        Advertiser_ID,
        Campaign,
        Insertion_Order,
        Device_Environment,
        Week_start
    ),

    subtotals AS
    (
      SELECT
        1 AS segment_number,
        segment1 AS segment,
        browser,
        SUM(imps) AS imps,
        SUM(safari_imps) AS safari_imps,
        Line_Item_Type,
        Advertiser,
        Advertiser_ID,
        Campaign,
        Insertion_Order,
        Device_Environment,
        Week_start
      FROM filtered
      WHERE segment1 IS NOT NULL
      GROUP BY 1, 2, 3, Line_Item_Type, Advertiser, Advertiser_ID, Campaign,Insertion_Order,Device_Environment, Week_start
      UNION ALL
      SELECT
        2 AS segment_number,
        segment2 AS segment,
        browser,
        SUM(imps) AS segment_browser_imps,
        SUM(safari_imps) AS safari_imps,
        Line_Item_Type,
        Advertiser,
        Advertiser_ID,
        Campaign,
        Insertion_Order,
        Device_Environment,
        Week_start
      FROM filtered
      WHERE Segment2 IS NOT NULL
      GROUP BY 1, 2, 3, Line_Item_Type, Advertiser, Advertiser_ID, Campaign,Insertion_Order,Device_Environment, Week_start
      UNION ALL
      SELECT
        3 AS segment_number,
        segment3 AS segment,
        browser,
        SUM(imps) AS segment_browser_imps,
        SUM(safari_imps) AS safari_imps,
        Line_Item_Type,
        Advertiser,
        Advertiser_ID,
        Campaign,
        Insertion_Order,
        Device_Environment,
        Week_start
      FROM filtered
      WHERE Segment3 IS NOT NULL
      GROUP BY 1, 2, 3, Line_Item_Type, Advertiser, Advertiser_ID, Campaign,Insertion_Order,Device_Environment,Week_start
    ),


    grand_total AS
    (
      SELECT
        segment_number,
        SUM(imps) AS imps
      FROM subtotals
      GROUP BY 1
    )

    SELECT
      segment_number,
      segment,
      SUM(sbt.imps) AS impressions,
      SUM(sbt.safari_imps) AS safari_impressions,
      sbt.Line_Item_Type as Line_Item_Type,
      sbt.browser as Browser,
      sbt.Advertiser as Advertiser,
      sbt.Advertiser_ID as Advertiser_ID,
      sbt.Campaign as Campaign,
      sbt.Insertion_Order as Insertion_Order,
      sbt.Device_Environment as Device_Environment,
      sbt.Week_start as Week_start
    FROM `subtotals` sbt
    LEFT OUTER JOIN `grand_total` AS gt USING (segment_number)
    GROUP BY 1, 2, Line_Item_Type, Browser, Advertiser, Advertiser_ID, Campaign,Insertion_Order,Device_Environment,Week_start
    ORDER BY 1, 2
  """

  clean_browser_report = """
    SELECT
      DV3_reporting.Partner AS Partner_clean,
      DV3_reporting.Partner_Id,
      CONCAT(DV3_reporting.Partner," - ",CAST(DV3_reporting.Partner_Id AS STRING)) AS Partner,
      DV3_reporting.Advertiser AS Advertiser_clean,
      DV3_reporting.Advertiser_Id,
      CONCAT(DV3_reporting.Advertiser," - ",CAST(DV3_reporting.Advertiser_Id AS STRING)) AS Advertiser,
      DV3_reporting.Advertiser_Currency,
      DV3_reporting.Insertion_Order AS Insertion_Order_clean,
      DV3_reporting.Insertion_Order_Id,
      CONCAT(DV3_reporting.Insertion_Order," - ",CAST(DV3_reporting.Insertion_Order_Id AS STRING)) AS Insertion_Order,
      DV3_reporting.Campaign AS Campaign_clean,
      DV3_reporting.Campaign_Id,
      CONCAT(DV3_reporting.Campaign," - ",CAST(DV3_reporting.Campaign_Id AS STRING)) AS Campaign,
      DV3_reporting.Line_Item AS Line_Item_clean,
      DV3_reporting.Line_Item_Id,
      CONCAT(DV3_reporting.Line_Item," - ",CAST(DV3_reporting.Line_Item AS STRING)) AS Line_Item,
      DV3_reporting.Browser AS Browser_Detail,
      Bro.Browser_Platform_clean AS Browser,
      DV3_reporting.Environment AS Environment_Detail,
      Env.Environment_clean AS Environment,
      Dev.Device_Type AS Device_Type,
      Dev.Device AS Device,
      seg.Segment1 AS Segment1,
      seg.Segment2 AS Segment2,
      seg.Segment3 AS Segment3,
      DV3_reporting.Week,
      CAST(PARSE_DATE('%Y/%m/%d', SPLIT(DV3_reporting.Week,'-')[OFFSET(0)]) AS DATE) AS Week_start,
      DV3_reporting.Month,
      DV3_reporting.Year,
      DV3_reporting.Line_Item_Type,
      SUM(DV3_reporting.Impressions) AS Impressions,
      SUM(DV3_reporting.Clicks) AS Clicks,
      SUM(DV3_reporting.Total_Conversions) AS Total_Conversions,
      SUM(DV3_reporting.Post_Click_Conversions) AS Post_Click_Conversions,
      SUM(DV3_reporting.Post_View_Conversions) AS Post_View_Conversions,
      SUM(DV3_reporting.CM_Post_View_Revenue) AS CM_Post_View_Revenue,
      SUM(DV3_reporting.CM_Post_Click_Revenue) AS CM_Post_Click_Revenue,
      SUM(DV3_reporting.Revenue_Adv_Currency) AS Revenue_Adv_Currency,
      SUM(DV3_reporting.Media_Cost_Advertiser_Currency) AS Media_Cost_Advertiser_Currency,
      SAFE_DIVIDE(SUM(DV3_reporting.Revenue_Adv_Currency), SAFE_DIVIDE(SUM(DV3_reporting.Impressions), 1000)) AS CPM,
      SAFE_DIVIDE(SUM(DV3_reporting.Revenue_Adv_Currency), SUM(DV3_reporting.Total_Conversions)) AS CPA,
      SAFE_DIVIDE(SUM(DV3_reporting.CM_Post_View_Revenue + DV3_reporting.CM_Post_Click_Revenue), SUM(DV3_reporting.Revenue_Adv_Currency)) AS ROAS_Total,
      SAFE_DIVIDE(SUM(DV3_reporting.CM_Post_View_Revenue), SUM(DV3_reporting.Revenue_Adv_Currency)) AS ROAS_View,
      SAFE_DIVIDE(SUM(DV3_reporting.CM_Post_Click_Revenue), SUM(DV3_reporting.Revenue_Adv_Currency)) AS ROAS_Click
    FROM
      `{dataset}.z_Dv360_Browser_Report_Dirty` AS DV3_reporting
    LEFT JOIN
      `{dataset}.z_Browser` AS Bro
    ON
      DV3_reporting.Browser = Bro.Browser_Platform
    LEFT JOIN
      `{dataset}.z_Environment` AS Env
    ON
      DV3_reporting.Environment = Env.Environment
    LEFT JOIN
      `{dataset}.z_Device_Type` AS Dev
    ON
      DV3_reporting.Device_Type = Dev.Device_Type
    LEFT JOIN
      `{dataset}.z_Custom_Segments` AS seg
    ON
      DV3_reporting.Line_Item_Id = seg.Line_Item_Id
    GROUP BY
      DV3_reporting.Partner,
      DV3_reporting.Partner_Id,
      DV3_reporting.Advertiser,
      DV3_reporting.Advertiser_Id,
      DV3_reporting.Advertiser_Currency,
      DV3_reporting.Insertion_Order,
      DV3_reporting.Insertion_Order_Id,
      DV3_reporting.Campaign,
      DV3_reporting.Campaign_Id,
      DV3_reporting.Line_Item,
      DV3_reporting.Line_Item_Id,
      DV3_reporting.Browser,
      Bro.Browser_Platform_clean,
      DV3_reporting.Environment,
      Env.Environment_clean,
      DV3_reporting.Week,
      DV3_reporting.Month,
      DV3_reporting.Year,
      DV3_reporting.Line_Item_Type,
      Dev.Device,
      Device_Type,
      Week_start,
      seg.Segment1,
      seg.Segment2,
      seg.Segment3
  """

  cm_floodlight_join = """
    SELECT
      Flood.Floodlight_Attribution_Type AS Floodlight_Attribution_Type,
      Att.Attribution_Type AS Attribution_Type,
      CMBrowser.Browser_Platform AS Browser_Platform,
      CMBrowser.Browser_Platform_detail AS Browser_Platform_detail,
      CMBrowser.Browser_Platform_clean AS Browser_Platform_clean,
      SUM(Total_Conversions) AS Total_Conversions,
      SUM(Click_Through_Conversions) AS Click_Through_Conversions,
      SUM(View_Through_Conversions) AS View_Through_Conversions
    FROM
      `{dataset}.z_Floodlight_CM_Report` AS Flood
    JOIN
      `{dataset}.z_Floodlight_Attribution` AS Att
    ON
      Flood.Floodlight_Attribution_Type = Att.Floodlight_Attribution_Type
    LEFT JOIN
      `{dataset}.z_CM_Browser_lookup` AS CMBrowser
    ON
      Flood.Browser_Platform = CMBrowser.Browser_Platform
    GROUP BY
      Flood.Floodlight_Attribution_Type,
      Att.Attribution_Type,
      Browser_Platform,
      Browser_Platform_detail,
      Browser_Platform_clean
  """

  cm_floodlight_multichart = """
    WITH
    attrtype_browser_total AS
    (
      SELECT
        IFNULL(att.Attribution_Type, '[missing]') as attribution_type,
        IFNULL(cm_br.Browser_Platform_clean, '[missing]') as browser_platform,
        SUM(fl.Total_Conversions) as convs
      FROM `{dataset}.z_CM_Floodlight` fl
      LEFT OUTER JOIN `{dataset}.z_Floodlight_Attribution` AS att USING(Floodlight_Attribution_Type)
      LEFT OUTER JOIN `{dataset}.z_CM_Browser_lookup` AS cm_br USING(Browser_Platform_clean)
      WHERE cm_br.Browser_Platform_clean IN ('Chrome', 'Safari', 'FF', 'MSFT')
      GROUP BY 1, 2
    ),


    grand_total AS
    (
      SELECT SUM(convs) AS convs
      FROM attrtype_browser_total
    )

    SELECT
      browser_platform,
      SAFE_DIVIDE(SUM(IF(UPPER(attribution_type)='ATTRIBUTED', abt.convs, 0)),
                  SUM(abt.convs)) AS percent_attributed,
      SAFE_DIVIDE(SUM(IF(UPPER(attribution_type)='UNATTRIBUTED', abt.convs, 0)),
                  SUM(abt.convs)) AS percent_unattributed,
      SAFE_DIVIDE(SUM(abt.convs), ANY_VALUE(gt.convs)) share_of_floodlight_conversions
    FROM attrtype_browser_total abt
    CROSS JOIN grand_total gt
    GROUP BY 1
    ORDER BY 4 DESC
  """

  cm_segmentation = """
    SELECT
      CONCAT(CM.Advertiser," - ",CAST(CM.Advertiser_Id AS STRING)) AS Advertiser,
      CONCAT(CM.Campaign," - ",CAST(CM.Campaign_Id AS STRING)) AS Campaign,
      CM.Site_Dcm,
      CM.Browser_Platform AS Browser_Platform,
      CMBrowser.Browser_Platform_detail AS Browser_Platform_detail,
      CMBrowser.Browser_Platform_clean AS Browser_Platform_clean,
      Platform_Type,
      SiteSeg.Site_Type AS Site_Type,
      CAST(Week AS DATE) AS Week,
      SUM(CM.Impressions) AS Impressions,
      SUM(Clicks) AS Clicks,
      SUM(Total_Conversions) AS Total_Conversions,
      SUM(Click_Through_Conversions) AS Click_Through_Conversions,
      SUM(View_Through_Conversions) AS View_Through_Conversions
    FROM
      `{dataset}.z_CM_Browser_Report_Dirty` AS CM
    LEFT JOIN
      `{dataset}.z_CM_Site_Segmentation` AS SiteSeg
    ON
      CM.Site_Dcm = SiteSeg.Site_Dcm
    LEFT JOIN
      `{dataset}.z_CM_Browser_lookup` AS CMBrowser
    ON
      CM.Browser_Platform = CMBrowser.Browser_Platform
    GROUP BY
      Advertiser,
      Campaign,
      Site_Dcm,
      Browser_Platform,
      Browser_Platform_detail,
      Browser_Platform_clean,
      Platform_Type,
      Site_Type,
      Week
  """

  cm_site_segmentation = """
    SELECT
    r.Site_Dcm,
    Sum(r.Impressions) AS Impressions,
    s.Site_Type
    FROM `{dataset}.z_CM_Browser_Report_Dirty` as r
    left join `{dataset}.z_CM_Site_Segmentation_Sheet` as s
    on r.Site_Dcm = s.Site_Dcm
    Group By
    Site_Dcm,
    Site_Type
    Order By
    Impressions desc
  """

  dv360_custom_segments = """
    SELECT DISTINCT
      c.Advertiser,
      c.Advertiser_Id,
      c.Campaign,
      c.Campaign_Id,
      c.Insertion_Order,
      c.Insertion_Order_Id,
      c.Line_Item,
      c.Line_Item_Id,
      c.Line_Item_Type,
      sum(c.Impressions),
      s.Segment1,
      s.Segment2,
      s.Segment3
    from
      `{dataset}.z_Dv360_Browser_Report_Dirty` as c
    left join
      `{dataset}.z_Custom_Segments_Sheet` as s
    on
      c.Line_Item_Id = s.Line_Item_Id
    where
      c.Line_Item_Type != "TrueView"
    Group By
      c.Advertiser,
      c.Advertiser_Id,
      c.Campaign,
      c.Campaign_Id,
      c.Insertion_Order,
      c.Insertion_Order_Id,
      c.Line_Item,
      c.Line_Item_Id,
      c.Line_Item_Type,
      s.Segment1,
      s.Segment2,
      s.Segment3
  """

  safari_distribution_90days = """
      SELECT
      Partner,
      Advertiser,
      Campaign,
      Environment,
      Device,
      CASE
        when Line_Item_Type = "TrueView" THEN "TrueView"
        ELSE Concat(Device,"-",Environment)
      END AS Device_Environment,
      Week_start,
      Line_item_type,
      CASE
        WHEN Browser_detail="Safari" THEN "Safari 12+13"
        WHEN Browser_detail="Safari 12" THEN "Safari 12+13"
        WHEN Browser_detail="Safari 11" THEN "Safari 11"
      ELSE
      "Other Safari"
      END
        AS Browser_Rollup,
        SUM(Impressions) AS Impressions
      FROM
        `{dataset}.z_DV360_Browser_Report_Clean`
      WHERE
        Browser = 'Safari'
        AND
        DATE_DIFF(CURRENT_DATE(),Week_start,WEEK)<12
      GROUP BY
        Browser_Rollup,
        Partner,
        Advertiser,
        Campaign,
        Environment,
        Device,
        Device_Environment,
        Line_item_type,
        Week_start
      ORDER BY
        Impressions DESC
  """

  sdf_join = """
    SELECT
    C.Campaign_Id as Campaign_Id,
    C.Advertiser_Id as Advertiser_Id,
    C.Name as Campaign_Name,
    IO.Io_Id as IO_Id,
    IO.Name as IO_Name,
    IO.Io_Type as IO_Type,
    IO.Pacing as IO_Pacing,
    IO.Performance_Goal_Type as IO_Performance_Goal_Type,
    IO.Performance_Goal_Value as IO_Performance_Goal_Value,
    IO.Budget_Type as IO_Budget_Type,
    IO.Budget_Segments as IO_Budget_Segments,
    IO.Auto_Budget_Allocation as IO_Auto_Budget_Allocation,
    LI.Line_Item_Id as LI_Id,
    LI.Type as LI_Type,
    LI.Subtype as LI_Subtype,
    LI.Name as LI_Name,
    LI.Budget_Type as LI_Budget_Type,
    LI.Budget_Amount as LI_Budget_Amount,
    LI.Pacing as LI_Pacing,
    LI.Pacing_Rate as LI_Pacing_Rate,
    LI.Bid_Strategy_Type as LI_Bid_Strategy_Type,
    LI.Bid_Strategy_Do_Not_Exceed as LI_Bid_Strategy_Do_Not_Exceed,
    LI.Geography_Targeting_Include as LI_Geography_Targeting_Include,
    LI.Geography_Targeting_Exclude as LI_Geography_Targeting_Exclude,
    LI.Device_Targeting_Include as LI_Device_Targeting_Include,
    LI.Device_Targeting_Exclude as LI_Device_Targeting_Exclude,
    LI.Browser_Targeting_Include as LI_Browser_Targeting_Include,
    LI.Browser_Targeting_Exclude as LI_Browser_Targeting_Exclude,

    LI.Third_Party_Verification_Services as LI_Third_Party_Verification_Services,
    LI.Third_Party_Verification_Labels as LI_Third_Party_Verification_Labels,
    LI.Channel_Targeting_Include as LI_Channel_Targeting_Include,
    LI.Channel_Targeting_Exclude as LI_Channel_Targeting_Exclude,
    LI.Site_Targeting_Include as LI_Site_Targeting_Include,
    LI.Site_Targeting_Exclude as LI_Site_Targeting_Exclude,
    LI.App_Targeting_Include as LI_App_Targeting_Include,
    LI.App_Targeting_Exclude as LI_App_Targeting_Exclude,
    LI.App_Collection_Targeting_Include as LI_App_Collection_Targeting_Include,
    LI.App_Collection_Targeting_Exclude as LI_App_Collection_Targeting_Exclude,
    LI.Category_Targeting_Include as LI_Category_Targeting_Include,
    LI.Category_Targeting_Exclude as LI_Category_Targeting_Exclude,
    LI.Keyword_Targeting_Include as LI_Keyword_Targeting_Include,
    LI.Keyword_Targeting_Exclude as LI_Keyword_Targeting_Exclude,
    LI.Keyword_List_Targeting_Exclude as LI_Keyword_List_Targeting_Exclude,
    LI.Audience_Targeting_Similar_Audiences as LI_Audience_Targeting_Similar_Audiences,
    LI.Audience_Targeting_Include as LI_Audience_Targeting_Include,
    LI.Audience_Targeting_Exclude as LI_Audience_Targeting_Exclude,
    LI.Affinity_In_Market_Targeting_Include as LI_Affinity_In_Market_Targeting_Include,
    LI.Affinity_In_Market_Targeting_Exclude as LI_Affinity_In_Market_Targeting_Exclude,
    LI.Custom_List_Targeting as LI_Custom_List_Targeting,
    LI.Daypart_Targeting as LI_Daypart_Targeting,
    LI.Daypart_Targeting_Time_Zone as LI_Daypart_Targeting_Time_Zone,
    LI.Environment_Targeting as LI_Environment_Targeting,
    LI.Demographic_Targeting_Gender as LI_Demographic_Targeting_Gender,
    LI.Demographic_Targeting_Age as LI_Demographic_Targeting_Age,
    LI.Demographic_Targeting_Household_Income as LI_Demographic_Targeting_Household_Income,
    LI.Demographic_Targeting_Parental_Status as LI_Demographic_Targeting_Parental_Status

    FROM
      `{dataset}.SDF_campaign` AS C
    LEFT JOIN
      `{dataset}.SDF_insertion_order` AS IO
    ON
      C.Campaign_Id = IO.Campaign_Id
    LEFT JOIN
      `{dataset}.SDF_line_item` AS LI
    ON
      IO.Io_Id = LI.Io_Id
    GROUP BY
    Campaign_Id,
    Advertiser_Id,
    Campaign_Name,
    IO_Id,
    IO_Name,
    IO_Type,
    IO_Pacing,
    IO_Performance_Goal_Type,
    IO_Performance_Goal_Value,
    IO_Budget_Type,
    IO_Budget_Segments,
    IO_Auto_Budget_Allocation,
    LI_Id,
    LI_Type,
    LI_Subtype,
    LI_Name,
    LI_Budget_Type,
    LI_Budget_Amount,
    LI_Pacing,
    LI_Pacing_Rate,
    LI_Bid_Strategy_Type,
    LI_Bid_Strategy_Do_Not_Exceed,
    LI_Geography_Targeting_Include,
    LI_Geography_Targeting_Exclude,
    LI_Device_Targeting_Include,
    LI_Device_Targeting_Exclude,
    LI_Browser_Targeting_Include,
    LI_Browser_Targeting_Exclude,
    LI_Third_Party_Verification_Services,
    LI_Third_Party_Verification_Labels,
    LI_Channel_Targeting_Include,
    LI_Channel_Targeting_Exclude,
    LI_Site_Targeting_Include,
    LI_Site_Targeting_Exclude,
    LI_App_Targeting_Include,
    LI_App_Targeting_Exclude,
    LI_App_Collection_Targeting_Include,
    LI_App_Collection_Targeting_Exclude,
    LI_Category_Targeting_Include,
    LI_Category_Targeting_Exclude,
    LI_Keyword_Targeting_Include,
    LI_Keyword_Targeting_Exclude,
    LI_Keyword_List_Targeting_Exclude,
    LI_Audience_Targeting_Similar_Audiences,
    LI_Audience_Targeting_Include,
    LI_Audience_Targeting_Exclude,
    LI_Affinity_In_Market_Targeting_Include,
    LI_Affinity_In_Market_Targeting_Exclude,
    LI_Custom_List_Targeting,
    LI_Daypart_Targeting,
    LI_Daypart_Targeting_Time_Zone,
    LI_Environment_Targeting,
    LI_Demographic_Targeting_Gender,
    LI_Demographic_Targeting_Age,
    LI_Demographic_Targeting_Household_Income,
    LI_Demographic_Targeting_Parental_Status
  """



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os
import re

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from starthinker_ui.recipe.scripts import Script
from website.views import solutions, solution, help

RE_TRAILING_WHITESPACE = re.compile(r'\s+$', re.MULTILINE)

class Command(BaseCommand):
  help = 'Generate HTML For GitHub'

  def add_arguments(self, parser):
    parser.add_argument(
        '--analytics',
        action='store',
        dest='analytics',
        required=True,
        type=str,
        help='Google analytics tag.',
    )

  def strip_trailing_whitespace(self, value):
    return RE_TRAILING_WHITESPACE.sub('', value) + '\n'

  def handle(self, *args, **kwargs):

    settings.GOOGLE_ANALYTICS = kwargs['analytics']

    directory = '%s/docs' % settings.UI_ROOT
    print('Writing:', directory)
    with open('%s/index.html' % directory, 'w') as index_file:
      index_file.write(self.strip_trailing_whitespace(solutions(request=None)))

    directory = '%s/docs/help' % settings.UI_ROOT
    print('Writing:', directory)
    os.makedirs(directory, exist_ok=True)
    with open('%s/index.html' % directory, 'w') as index_file:
      index_file.write(self.strip_trailing_whitespace(help(request=None)))

    directory = '%s/docs/solution' % settings.UI_ROOT
    print('Writing:', directory)
    os.makedirs(directory, exist_ok=True)
    with open('%s/index.html' % directory, 'w') as index_file:
      index_file.write(self.strip_trailing_whitespace(solutions(request=None)))

    for s in Script.get_scripts():
      if s.get_open_source():
        directory = '%s/docs/solution/%s' % (settings.UI_ROOT, s.get_tag())
        print('Writing:', directory)
        os.makedirs(directory, exist_ok=True)
        with open('%s/index.html' % directory, 'w') as solution_file:
          solution_file.write(self.strip_trailing_whitespace(solution(request=None, tag=s.get_tag())))



from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from starthinker_ui.recipe.scripts import Script
from starthinker_ui.recipe.colab import script_to_colab


class Command(BaseCommand):
  help = 'Generate Templates For Colab'

  def handle(self, *args, **kwargs):

    for script in Script.get_scripts():
      print('%s, %s, "%s", %s' %
            (script.get_name(), script.get_requirements(),
             script.get_description().replace('"', '\''), script.get_authors()))



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from datetime import date, timedelta

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from starthinker.util.email import send_email
from starthinker.util.email.template import EmailTemplate
from starthinker_ui.recipe.scripts import Script


class Command(BaseCommand):
  help = 'Generate Newsletter Of New Solutions'

  def add_arguments(self, parser):
    parser.add_argument(
        '--days',
        action='store',
        dest='days',
        default=90,
        type=int,
        help='Number of days to go back for recipes.',
    )

  def handle(self, *args, **kwargs):
    day = date.today() - timedelta(days=kwargs['days'])

    email = {
        'subject':
            'Announcing Six New Open Source Modules For Ecosystems',
        'style': {
            'background': '#f2f2f2',
            'foreground': '#ffffff',
            'text': '#414347',
            'link': '#4285f4',
            'font': 'Roboto, Helvetica, Arial sans-serif;',
            'align': 'left'
        },
        'logo':
            'https://storage.googleapis.com/starthinker-ui/gTech_StarThinker.png',
        'body': {
            'sections': [{
                'header':
                    'Six New Solutions For Partners To Build New Services',
                'paragraph':
                    'In Q1, StarThinker released 6 new building blocks '
                    'available as Python, Airflow, Colab, and no-coding UI.  '
                    'These building blocks are now open sourve and availbale '
                    'for deployment by Partners.  Below is a description of '
                    'each solution and possible service or efficiency gain by '
                    'partners.',
                'grid': []
            }]
        },
        'footer': [{
            'text': 'Internal UI',
            'link': 'http://go/starthinker'
        }, {
            'text': 'Esteemed Solution Gallery',
            'link': 'http://go/esteemed/#/search/StarThinker'
        }, {
            'text': 'GitHub Solution Gallery',
            'link': 'https://google.github.io/starthinker/'
        }, {
            'text': 'Google3 Repository',
            'link': 'http://go/starthinker-google3'
        }, {
            'text': 'GOB Repository ( Official )',
            'link': 'http://go/starthinker-code'
        }, {
            'text': 'GitHub Repository',
            'link': 'https://github.com/google/starthinker'
        }],
        'copyright':
            'Copyright 2020 Google LLC'
    }

    odd = True
    for s in Script.get_scripts():
      if s.get_released() < day:
        continue
      print('SCRIPT: ', s.get_tag())

      if not s.get_image():
        continue

      row = [{
          'image': {
              'src': s.get_image(),
              'link': s.get_link_client()
          }
      }, {
          'header': '[%s](%s)' % (s.get_name(), s.get_link_client()),
          'paragraph': s.get_description()
      }]
      email['body']['sections'][0]['grid'].append(row)

      if odd:
        row.reverse()
      odd = not odd

    email = EmailTemplate(email)

    # send or print
    #if project.args.email_to and project.args.email_from:
    #  print('EMAILING: ', project.args.email_to)
    #  send_email('user', project.args.email_to, project.args.email_from, None, email.get_subject(), email.get_text(), email.get_html())
    #else:
    if 1:
      # write to STDOUT
      print(email.get_html())
      print('<pre style="width:600px;margin:0px auto;">%s</pre>' %
            email.get_text())



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""WSGI config for ui project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/howto/deployment/wsgi/
"""

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'starthinker_ui.ui.settings')

application = get_wsgi_application()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

ITP Audit Dashboard ( 2020 )

Dashboard that shows performance metrics across browser to see the impact of ITP.

  - Follow the instructions from <a href="https://docs.google.com/document/d/1HaRCMaBBEo0tSKwnofWNtaPjlW0ORcVHVwIRabct4fY/edit?usp=sharing" target="_blank">this document</a>

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_timezone': 'America/Los_Angeles',  # Timezone for report dates.
  'recipe_name': '',  # Name of document to deploy to.
  'auth_write': 'service',  # Credentials used for writing data.
  'recipe_slug': 'ITP_Audit_Dashboard',  # BigQuery dataset for store dashboard tables.
  'auth_read': 'user',  # Credentials used for reading data.
  'cm_account_id': '',  # Campaign Manager Account Id.
  'date_range': 'LAST_365_DAYS',  # Timeframe to run the ITP report for.
  'cm_advertiser_ids': '',  # Optional: Comma delimited list of CM advertiser ids.
  'floodlight_configuration_id': '',  # Floodlight Configuration Id for the Campaign Manager floodlight report.
  'dv360_partner_ids': '',  # Comma delimited list of DV360 Partner ids.
  'dv360_advertiser_ids': '',  # Optional: Comma delimited list of DV360 Advertiser ids.
}

RECIPE = {
  'setup': {
    'hour': [
      3
    ],
    'day': [
      'Mon'
    ]
  },
  'tasks': [
    {
      'drive': {
        'auth': 'user',
        'hour': [
        ],
        'copy': {
          'source': 'https://docs.google.com/spreadsheets/d/1rH_PGXOYW2mVdmAYnKbv6kcaB6lQihAyMsGtFfinnqg/',
          'destination': {
            'field': {
              'name': 'recipe_name',
              'prefix': 'ITP Audit ',
              'kind': 'string',
              'order': 1,
              'description': 'Name of document to deploy to.',
              'default': ''
            }
          }
        }
      }
    },
    {
      'dataset': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'order': 1,
            'default': 'ITP_Audit_Dashboard',
            'description': 'BigQuery dataset for store dashboard tables.'
          }
        }
      }
    },
    {
      'dbm': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'report': {
          'timeout': 90,
          'filters': {
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'dv360_advertiser_ids',
                  'kind': 'integer_list',
                  'order': 6,
                  'default': '',
                  'description': 'Optional: Comma delimited list of DV360 Advertiser ids.'
                }
              }
            },
            'FILTER_PARTNER': {
              'values': {
                'field': {
                  'name': 'dv360_partner_ids',
                  'kind': 'integer_list',
                  'order': 5,
                  'default': '',
                  'description': 'Comma delimited list of DV360 Partner ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'ITP_Audit_Browser_',
                  'order': 1,
                  'description': 'Name of report in DV360, should be unique.'
                }
              },
              'dataRange': {
                'field': {
                  'name': 'date_range',
                  'kind': 'choice',
                  'order': 3,
                  'default': 'LAST_365_DAYS',
                  'choices': [
                    'LAST_7_DAYS',
                    'LAST_14_DAYS',
                    'LAST_30_DAYS',
                    'LAST_365_DAYS',
                    'LAST_60_DAYS',
                    'LAST_7_DAYS',
                    'LAST_90_DAYS',
                    'MONTH_TO_DATE',
                    'PREVIOUS_MONTH',
                    'PREVIOUS_QUARTER',
                    'PREVIOUS_WEEK',
                    'PREVIOUS_YEAR',
                    'QUARTER_TO_DATE',
                    'WEEK_TO_DATE',
                    'YEAR_TO_DATE'
                  ],
                  'description': 'Timeframe to run the ITP report for.'
                }
              },
              'format': 'CSV'
            },
            'params': {
              'type': 'TYPE_GENERAL',
              'groupBys': [
                'FILTER_PARTNER',
                'FILTER_PARTNER_NAME',
                'FILTER_ADVERTISER',
                'FILTER_ADVERTISER_NAME',
                'FILTER_ADVERTISER_CURRENCY',
                'FILTER_MEDIA_PLAN',
                'FILTER_MEDIA_PLAN_NAME',
                'FILTER_INSERTION_ORDER',
                'FILTER_INSERTION_ORDER_NAME',
                'FILTER_LINE_ITEM',
                'FILTER_LINE_ITEM_NAME',
                'FILTER_PAGE_LAYOUT',
                'FILTER_WEEK',
                'FILTER_MONTH',
                'FILTER_YEAR',
                'FILTER_LINE_ITEM_TYPE',
                'FILTER_DEVICE_TYPE',
                'FILTER_BROWSER'
              ],
              'metrics': [
                'METRIC_MEDIA_COST_ADVERTISER',
                'METRIC_IMPRESSIONS',
                'METRIC_CLICKS',
                'METRIC_TOTAL_CONVERSIONS',
                'METRIC_LAST_CLICKS',
                'METRIC_LAST_IMPRESSIONS',
                'METRIC_CM_POST_CLICK_REVENUE',
                'METRIC_CM_POST_VIEW_REVENUE',
                'METRIC_REVENUE_ADVERTISER'
              ]
            }
          }
        },
        'delete': False
      }
    },
    {
      'dcm': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'report': {
          'timeout': 90,
          'account': {
            'field': {
              'name': 'cm_account_id',
              'kind': 'string',
              'order': 2,
              'default': '',
              'description': 'Campaign Manager Account Id.'
            }
          },
          'body': {
            'kind': 'dfareporting#report',
            'name': {
              'field': {
                'name': 'recipe_name',
                'kind': 'string',
                'prefix': 'ITP_Audit_Floodlight_',
                'order': 1,
                'description': 'Name of report in DV360, should be unique.'
              }
            },
            'format': 'CSV',
            'type': 'FLOODLIGHT',
            'floodlightCriteria': {
              'dateRange': {
                'kind': 'dfareporting#dateRange',
                'relativeDateRange': 'LAST_30_DAYS'
              },
              'floodlightConfigId': {
                'kind': 'dfareporting#dimensionValue',
                'dimensionName': 'dfa:floodlightConfigId',
                'value': {
                  'field': {
                    'name': 'floodlight_configuration_id',
                    'kind': 'integer',
                    'order': 4,
                    'default': '',
                    'description': 'Floodlight Configuration Id for the Campaign Manager floodlight report.'
                  }
                },
                'matchType': 'EXACT'
              },
              'reportProperties': {
                'includeUnattributedIPConversions': False,
                'includeUnattributedCookieConversions': True
              },
              'dimensions': [
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:site'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:floodlightAttributionType'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:interactionType'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:pathType'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:browserPlatform'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:platformType'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:week'
                }
              ],
              'metricNames': [
                'dfa:activityClickThroughConversions',
                'dfa:activityViewThroughConversions',
                'dfa:totalConversions',
                'dfa:totalConversionsRevenue'
              ]
            },
            'schedule': {
              'active': True,
              'repeats': 'WEEKLY',
              'every': 1,
              'repeatsOnWeekDays': [
                'Sunday'
              ]
            },
            'delivery': {
              'emailOwner': False
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': 'ITP_Audit_Dashboard',
                'description': 'BigQuery dataset for store dashboard tables.'
              }
            },
            'table': 'z_Floodlight_CM_Report',
            'is_incremental_load': False
          }
        },
        'delete': False
      }
    },
    {
      'dcm': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'report': {
          'timeout': 90,
          'account': {
            'field': {
              'name': 'cm_account_id',
              'kind': 'string',
              'order': 2,
              'default': '',
              'description': 'Campaign Manager Account Id.'
            }
          },
          'filters': {
            'dfa:advertiser': {
              'values': {
                'field': {
                  'name': 'cm_advertiser_ids',
                  'kind': 'integer_list',
                  'order': 3,
                  'default': '',
                  'description': 'Optional: Comma delimited list of CM advertiser ids.'
                }
              }
            }
          },
          'body': {
            'kind': 'dfareporting#report',
            'name': {
              'field': {
                'name': 'recipe_name',
                'kind': 'string',
                'order': 1,
                'prefix': 'ITP_Audit_Browser_',
                'default': 'ITP_Audit_Dashboard_Browser',
                'description': 'Name of the Campaign Manager browser report.'
              }
            },
            'fileName': {
              'field': {
                'name': 'recipe_name',
                'kind': 'string',
                'order': 1,
                'prefix': 'ITP_Audit_Browser_',
                'default': 'ITP_Audit_Dashboard_Browser',
                'description': 'Name of the Campaign Manager browser report.'
              }
            },
            'format': 'CSV',
            'type': 'STANDARD',
            'criteria': {
              'dateRange': {
                'kind': 'dfareporting#dateRange',
                'relativeDateRange': {
                  'field': {
                    'name': 'date_range',
                    'kind': 'choice',
                    'order': 3,
                    'default': 'LAST_365_DAYS',
                    'choices': [
                      'LAST_7_DAYS',
                      'LAST_14_DAYS',
                      'LAST_30_DAYS',
                      'LAST_365_DAYS',
                      'LAST_60_DAYS',
                      'LAST_7_DAYS',
                      'LAST_90_DAYS',
                      'MONTH_TO_DATE',
                      'PREVIOUS_MONTH',
                      'PREVIOUS_QUARTER',
                      'PREVIOUS_WEEK',
                      'PREVIOUS_YEAR',
                      'QUARTER_TO_DATE',
                      'WEEK_TO_DATE',
                      'YEAR_TO_DATE'
                    ],
                    'description': 'Timeframe to run the ITP report for.'
                  }
                }
              },
              'dimensions': [
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:campaign'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:campaignId'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:site'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:advertiser'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:advertiserId'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:browserPlatform'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:platformType'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:month'
                },
                {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:week'
                }
              ],
              'metricNames': [
                'dfa:impressions',
                'dfa:clicks',
                'dfa:totalConversions',
                'dfa:activityViewThroughConversions',
                'dfa:activityClickThroughConversions'
              ],
              'dimensionFilters': [
              ]
            },
            'schedule': {
              'active': True,
              'repeats': 'WEEKLY',
              'every': 1,
              'repeatsOnWeekDays': [
                'Sunday'
              ]
            },
            'delivery': {
              'emailOwner': False
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': 'ITP_Audit_Dashboard',
                'description': 'BigQuery dataset for store dashboard tables.'
              }
            },
            'table': 'z_CM_Browser_Report_Dirty',
            'is_incremental_load': False
          }
        },
        'delete': False
      }
    },
    {
      'sheets': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'sheet': {
          'field': {
            'name': 'recipe_name',
            'prefix': 'ITP Audit ',
            'kind': 'string',
            'order': 1,
            'description': 'Name of document to deploy to.',
            'default': ''
          }
        },
        'tab': 'Enviroment',
        'range': 'A:B',
        'header': True,
        'out': {
          'auth': {
            'field': {
              'name': 'auth_write',
              'kind': 'authentication',
              'order': 1,
              'default': 'service',
              'description': 'Credentials used for writing data.'
            }
          },
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': 'ITP_Audit_Dashboard',
                'description': 'BigQuery dataset for store dashboard tables.'
              }
            },
            'table': 'z_Environment'
          }
        }
      }
    },
    {
      'sheets': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'sheet': {
          'field': {
            'name': 'recipe_name',
            'prefix': 'ITP Audit ',
            'kind': 'string',
            'order': 1,
            'description': 'Name of document to deploy to.',
            'default': ''
          }
        },
        'tab': 'Browser',
        'range': 'A:C',
        'header': True,
        'out': {
          'auth': {
            'field': {
              'name': 'auth_write',
              'kind': 'authentication',
              'order': 1,
              'default': 'service',
              'description': 'Credentials used for writing data.'
            }
          },
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': 'ITP_Audit_Dashboard',
                'description': 'BigQuery dataset for store dashboard tables.'
              }
            },
            'table': 'z_Browser'
          }
        }
      }
    },
    {
      'sheets': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'sheet': {
          'field': {
            'name': 'recipe_name',
            'prefix': 'ITP Audit ',
            'kind': 'string',
            'order': 1,
            'description': 'Name of document to deploy to.',
            'default': ''
          }
        },
        'tab': 'CM_Browser_lookup',
        'range': 'A:C',
        'header': True,
        'out': {
          'auth': {
            'field': {
              'name': 'auth_write',
              'kind': 'authentication',
              'order': 1,
              'default': 'service',
              'description': 'Credentials used for writing data.'
            }
          },
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': 'ITP_Audit_Dashboard',
                'description': 'BigQuery dataset for store dashboard tables.'
              }
            },
            'table': 'z_CM_Browser_lookup'
          }
        }
      }
    },
    {
      'sheets': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'sheet': {
          'field': {
            'name': 'recipe_name',
            'prefix': 'ITP Audit ',
            'kind': 'string',
            'order': 1,
            'description': 'Name of document to deploy to.',
            'default': ''
          }
        },
        'tab': 'Device_Type',
        'range': 'A:B',
        'header': True,
        'out': {
          'auth': {
            'field': {
              'name': 'auth_write',
              'kind': 'authentication',
              'order': 1,
              'default': 'service',
              'description': 'Credentials used for writing data.'
            }
          },
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': 'ITP_Audit_Dashboard',
                'description': 'BigQuery dataset for store dashboard tables.'
              }
            },
            'table': 'z_Device_Type'
          }
        }
      }
    },
    {
      'sheets': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'sheet': {
          'field': {
            'name': 'recipe_name',
            'prefix': 'ITP Audit ',
            'kind': 'string',
            'order': 1,
            'description': 'Name of document to deploy to.',
            'default': ''
          }
        },
        'tab': 'Floodlight_Attribution',
        'range': 'A:B',
        'header': True,
        'out': {
          'auth': {
            'field': {
              'name': 'auth_write',
              'kind': 'authentication',
              'order': 1,
              'default': 'service',
              'description': 'Credentials used for writing data.'
            }
          },
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': 'ITP_Audit_Dashboard',
                'description': 'BigQuery dataset for store dashboard tables.'
              }
            },
            'table': 'z_Floodlight_Attribution'
          }
        }
      }
    },
    {
      'dbm': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'report': {
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'ITP_Audit_Browser_',
              'order': 1,
              'description': 'Name of report in DV360, should be unique.'
            }
          }
        },
        'out': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': 'ITP_Audit_Dashboard',
                'description': 'BigQuery dataset for store dashboard tables.'
              }
            },
            'table': 'z_Dv360_Browser_Report_Dirty',
            'header': True,
            'schema': [
              {
                'name': 'Partner_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Partner',
                'type': 'STRING'
              },
              {
                'name': 'Advertiser_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Advertiser',
                'type': 'STRING'
              },
              {
                'name': 'Advertiser_Currency',
                'type': 'STRING'
              },
              {
                'name': 'Campaign_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Campaign',
                'type': 'STRING'
              },
              {
                'name': 'Insertion_Order_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Insertion_Order',
                'type': 'STRING'
              },
              {
                'name': 'Line_Item_Id',
                'type': 'INTEGER'
              },
              {
                'name': 'Line_Item',
                'type': 'STRING'
              },
              {
                'name': 'Environment',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Week',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Month',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Year',
                'type': 'INTEGER',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Line_Item_Type',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Device_Type',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Browser',
                'type': 'STRING',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Media_Cost_Advertiser_Currency',
                'type': 'FLOAT',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Impressions',
                'type': 'INTEGER',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Clicks',
                'type': 'INTEGER',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Total_Conversions',
                'type': 'FLOAT',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Post_Click_Conversions',
                'type': 'FLOAT',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Post_View_Conversions',
                'type': 'FLOAT',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Cm_Post_Click_Revenue',
                'type': 'FLOAT',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Cm_Post_View_Revenue',
                'type': 'FLOAT',
                'mode': 'NULLABLE'
              },
              {
                'name': 'Revenue_Adv_Currency',
                'type': 'FLOAT',
                'mode': 'NULLABLE'
              }
            ]
          }
        }
      }
    },
    {
      'itp_audit': {
        'auth': 'service',
        'account': {
          'field': {
            'name': 'cm_account_id',
            'kind': 'string',
            'order': 2,
            'default': '',
            'description': 'Campaign Manager Account Id.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'order': 1,
            'default': 'ITP_Audit_Dashboard',
            'description': 'BigQuery dataset for store dashboard tables.'
          }
        },
        'sheet': {
          'field': {
            'name': 'recipe_name',
            'prefix': 'ITP Audit ',
            'kind': 'string',
            'order': 1,
            'description': 'Name of document to deploy to.',
            'default': ''
          }
        },
        'timeout': 60
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('itp_audit', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
from urllib.request import urlopen
from urllib.parse import urlencode

from starthinker.util.project import project
from starthinker.util.data import put_rows


def fred_series(api_key, frequency, **kwargs):
  kwargs['api_key'] = api_key
  kwargs['frequency'] = frequency
  kwargs['file_type'] = 'json'

  url = 'https://api.stlouisfed.org/fred/series/observations?' + urlencode(
      kwargs)
  for observation in json.loads(
      urlopen(url).read().decode('UTF-8'))['observations']:
    yield [
        observation['realtime_start'], observation['realtime_end'],
        observation['date'], observation['value']
    ]


def fred_regional(api_key, frequency, region, **kwargs):
  kwargs['api_key'] = api_key
  kwargs['frequency'] = frequency
  kwargs['region_type'] = region_type
  kwargs['file_type'] = 'json'
  kwargs['start_date'] = str(project.date)

  url = 'https://api.stlouisfed.org/geofred/regional/data?' + urlencode(kwargs)
  for observation in json.loads(
      urlopen(url).read().decode('UTF-8'))['observations']:
    yield [
        observation['region'], observation['code'], observation['date'],
        observation['series_id']
    ]


@project.from_parameters
def fred():

  if 'series' in project.task:

    for parameters in project.task['series']:

      name = 'FRED_SERIES_%s' % parameters['series_id']
      rows = fred_series(project.task['api_key'], project.task['frequency'],
                         **parameters)

      if 'bigquery' in project.task['out']:
        project.task['out']['bigquery']['schema'] = [{
            'name': 'realtime_start',
            'type': 'DATE',
            'mode': 'REQUIRED'
        }, {
            'name': 'realtime_end',
            'type': 'DATE',
            'mode': 'REQUIRED'
        }, {
            'name': 'day',
            'type': 'DATE',
            'mode': 'REQUIRED'
        }, {
            'name': 'value',
            'type': 'FLOAT',
            'mode': 'REQUIRED'
        }]

  elif 'regions' in project.task:
    for parameters in project.task['regions']:

      name = 'FRED_SERIES_%s' % parameters['series_group']
      rows = fred_regional(project.task['api_key'], project.task['frequency'],
                           project.task['region_type'], **parameters)

      if 'bigquery' in project.task['out']:
        project.task['out']['bigquery']['schema'] = [{
            'name': 'region',
            'type': 'STRING',
            'mode': 'REQUIRED'
        }, {
            'name': 'code',
            'type': 'INTEGER',
            'mode': 'REQUIRED'
        }, {
            'name': 'series_id',
            'type': 'STRING',
            'mode': 'REQUIRED'
        }, {
            'name': 'value',
            'type': 'FLOAT',
            'mode': 'REQUIRED'
        }]

  else:
    raise Excpetion(
        'MISSING CONFIGURATION: Specify either series_id or series_group.')

  put_rows(project.task['auth'], name, rows)


if __name__ == '__main__':
  fred()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django import forms

from starthinker_ui.recipe.scripts import Script
from starthinker_ui.recipe.models import Recipe
from starthinker_ui.recipe.forms import SetupForm
from starthinker_ui.recipe.forms_json import ScriptJsonForm


class ScriptForm(forms.Form):
  users = None

  def __init__(self, manual, recipe, account, *args, **kwargs):
    post = args[0] if len(args) > 0 else None

    # load scripts ( used when creating new recipe )
    if 'scripts' in kwargs:
      scripts = [s for s in kwargs.pop('scripts').split(',') if s]
    else:
      scripts = []

    # fetch the instance and load initial data
    self.instance = recipe or Recipe()
    self.setup = SetupForm(manual, account, post, instance=self.instance)
    super(ScriptForm, self).__init__(*args, **kwargs)

    # create blank forms
    self.blanks = []
    for s in Script.get_scripts(account.email):
      if s.is_manual() == manual:
        self.blanks.append(
            ScriptJsonForm('[BLANK]', s, {}, prefix='%s_[BLANK]' % s.get_tag()))

    # processing incoming form
    self.forms = []
    if post:
      for prefix in post.getlist('form_prefix'):
        tag, sequence = prefix.rsplit('_', 1)
        sequence = int(sequence) - 1
        s = Script(tag)
        self.forms.append(
            ScriptJsonForm(
                sequence,
                s,
                self.instance.get_values()[sequence]['values']
                if self.instance and sequence < len(self.instance.get_values())
                else {},
                post,
                prefix=prefix))

    # loading from existing recipe
    elif self.instance:
      for sequence, script in enumerate(self.instance.get_values()):
        s = Script(script['tag'])
        self.forms.append(
            ScriptJsonForm(
                sequence + 1,
                s,
                script['values'],
                prefix='%s_%d' % (s.get_tag(), sequence + 1)))

    # starting a new recipe
    else:
      for sequence, script in enumerate(scripts):
        s = Script(script)
        self.forms.append(
            ScriptJsonForm(
                sequence + 1,
                s, [],
                prefix='%s_%d' % (s.get_tag(), sequence + 1)))

  def is_valid(self):
    return super(ScriptForm, self).is_valid() & self.setup.is_valid() & all(
        [script.is_valid() for script in self.forms])

  def get_errors(self):
    errors = list(['%s: %s' % (k, v) for k, v in self.setup.errors.items()])
    for script in self.forms:
      errors.extend(['%s: %s' % (k, v) for k, v in script.errors.items()])
    return errors

  def get_scripts(self):
    scripts = [script.get_script() for script in self.forms
              ]  # fetch sequence and scripts
    return sorted([script for script in scripts if script],
                  key=lambda k: k['sequence']
                 )  # sort and return script ( if delete sript will be None )

  def save(self):
    self.instance = self.setup.save()
    self.instance.set_values(self.get_scripts())
    self.instance.save()
    self.instance.update()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Manages id maps and caching.

  Id mapping is important because if an execution fails, the context of maps of
  ext ids and concrete CM ids is lost, the Store maintains a persistent map that
  can be referenced througout an execution allowing for successful retries.

  Caching is important for performance reasons, and to reduce the number of
  calls to the CM API.
"""

#import json

#from starthinker.util.sheets import sheets_read, sheets_write, sheets_clear


class Store(object):
  """Class that handles interaction with the Store."""

  def __init__(self):
    """Initializes the store.

    Since this is a sigleton, before the fist usage the trix id and auth scheme
    must be set.
    """
    self._store = {}
    self._id_map = {}


#    self.trix_id = None
#    self.auth = None

#  def load_id_map(self):
#    """Loads the ID map from the Bulkdozer feed into the object.
#
#    """
#    if self.trix_id:
#      data = sheets_read(self.auth, self.trix_id, 'Store', 'A1:Z1')
#      content = ''
#      if data and data[0]:
#        for cell in data[0]:
#          content += cell
#
#        self._id_map = json.loads(content)
#      else:
#        self._id_map = {}

#  def save_id_map(self):
#    """Saves the ID map into the Bulkdozer feed.
#
#    """
#    if self.trix_id:
#      columns = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'W', 'X', 'Y', 'Z']
#
#      data = json.dumps(self._id_map)
#      data = [data[start:start+49999] for start in range(0, len(data), 49999)]
#      sheets_write(self.auth, self.trix_id, 'Store', 'A:' + columns[len(data) + 1], [data])

#  def clear(self):
#    """Clears the store in the Bulkdozer feed.
#
#    """
#    if self.trix_id:
#      sheets_clear(self.auth, self.trix_id, 'Store', 'A1:Z1')
#
#    self._store = {}
#    self._id_map = {}

  def map(self, entity, ext_id, dcm_id):
    """Maps a CM id and an ext id for an entity.

    Args:
      entity: The name of the entity for which the ID relates.
      ext_id: Placeholder ext id.
      dcm_id: Real CM id of the object.
    """
    if not entity in self._id_map:
      self._id_map[entity] = {}

    self._id_map[entity][ext_id] = dcm_id
    self._id_map[entity][dcm_id] = ext_id

  def translate(self, entity, identifier):
    """Given an id, returns its counterpart.

    ext id to cm id and vice versa.

    Args:
      entity: The name of the entity for which the ID relates.
      identifier: Ext id or actual CM id to map.
    """
    if entity in self._id_map and identifier in self._id_map[entity]:
      return self._id_map[entity][identifier]

    return None

  def set(self, entity, keys, item):
    """Sets an item in the cache.

    Args:
      entity: The name of the entity cache to use.
      keys: The keys to set this item to. Typically this will contain the ext id
        and the actual CM id.
      item: The item to cache.
    """
    if not entity in self._store:
      self._store[entity] = {}

    for key in keys:
      self._store[entity][str(key)] = item

  def get(self, entity, key):
    """Gets and item from the cache.

    Args:
      entity: The entity cache to use.
      key: The key to use to lookup the cached item.
    """
    if entity in self._store:
      return self._store[entity].get(str(key))

    return None

store = Store()
